#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File				: Ampel-core/ampel/t3/supply/select/T3FilteringStockSelector.py
# License			: BSD-3-Clause
# Author			: Jakob van Santen <jakob.van.santen@desy.de>
# Date				: 02.08.2020
# Last Modified Date:  02.08.2020
# Last Modified By	: Jakob van Santen <jakob.van.santen@desy.de>

import functools
import operator
from collections.abc import Generator
from itertools import islice
from typing import Any

from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.t3.T2FilterModel import T2FilterModel
from ampel.mongo.query.general import build_general_query
from ampel.mongo.utils import maybe_match_array
from ampel.t3.supply.select.T3StockSelector import T3StockSelector
from ampel.types import StockId


class T3FilteringStockSelector(T3StockSelector):
	"""
	Selector subclass that filters stocks based on their latest T2 results.
	
	Example configuration::
		
		{
			...
			"t2_filter": [
				{
					"unit": "T2SNCosmo",
					"match": {
						"fit_acceptable": True,
						"fit_results.c": {"$gt": 1},
					},
				}
			]
		}
	"""

	t2_filter: T2FilterModel | AllOf[T2FilterModel] | AnyOf[T2FilterModel]
	chunk_size: int = 200

	# Override/Implement
	def fetch(self) -> Generator[dict[str, Any], None, None]: # type: ignore[override]

		# Execute query on T0 collection to get target stocks
		if not (cursor := super().fetch()):
			return None

		# Execute aggregation on T2 collection to get matching subset of stocks
		# NB: aggregate in chunks to avoid the 100 MB aggregation memory limit
		input_count = 0
		output_count = 0
		while (stock_ids := [doc['stock'] for doc in islice(cursor, self.chunk_size)]):
			count = 0
			for count, doc in enumerate( # noqa: B007
				self.context.db.get_collection('t2').aggregate(
					self._t2_filter_pipeline(stock_ids),
				),
				1
			):
				yield doc

			input_count += len(stock_ids)
			output_count += count

		self.logger.info(f"{output_count}/{input_count} stocks passed filter criteria")



	def _build_match(self, f: T2FilterModel | AllOf[T2FilterModel] | AnyOf[T2FilterModel]) -> dict[str, Any]:
		if isinstance(f, T2FilterModel):
			return {f"{f.unit}.{k}": v for k, v in f.match.items()}
		if isinstance(f, AllOf):
			return {'$and': [self._build_match(el) for el in f.all_of]}
		if isinstance(f, AnyOf):
			return {'$or': [self._build_match(el) for el in f.any_of]}
		raise TypeError()


	def _t2_filter_pipeline(self, stock_ids: list[StockId]) -> list[dict]:
		return [
			*self._t2_merge_pipeline(stock_ids),
			{'$match': self._build_match(self.t2_filter)},
			{'$replaceRoot': {'newRoot': {'stock': '$_id'}}},
		]


	def _t2_merge_pipeline(self, stock_ids: list[StockId]) -> list[dict[str, Any]]:
		"""
		Create a pipeline for the T2 collection that yields docs whose _id is
		the stock id and whose remaining fields are the latest result for each
		T2. For example, given the following T2 docs::
			
			{
				'_id': 1,
				...
				'stock': 42,
				'unit': 'Unit1',
				'code': 0,
				'body': [{
					...
					'data': {'thing1': 3}
				}]
			},
			{
				'_id': 2,
				...
				'stock': 42,
				'unit': 'Unit2',
				'code': 0,
				'body': [{
					...
					'data': {'thing2': 7}
				}]
			}
		
		the pipeline yields a single doc of the form::
			
			{
				'_id': 42,
				'Unit1': {'thing1': 3},
				'Unit2': {'thing2': 7}
			}
		"""
		# NB: we reuse the general query here to ensure that we only process
		# T2s associated with the requested channels
		match = {
			'code': 0,
			'unit': maybe_match_array(_all_units(self.t2_filter)),
			**build_general_query(stock_ids, self.channel, self.tag)
		}

		return [
			# select t2 docs for target stocks
			{
				'$match': match
			},
			# find "latest" result for each stock and unit, where "latest" is
			# the document with the largest _id (creation time)
			{
				'$sort': {
					'stock': 1,
					'unit': 1,
					'_id': 1
				}
			},
			{'$unwind': '$body'},
			{
				'$group': {
					'_id': {
						'stock': '$stock',
						'unit': '$unit'
					},
					'data': {'$last': '$body'}
				}
			},
			# nest data under key named for unit, e.g.
			# {'unit': 'T2Unit', 'data': {'foo': 1}} -> {'T2Unit': {'foo': 1}}
			{
				'$replaceRoot': {
					'newRoot': {
						'_id': '$_id',
						'data': {
							'$arrayToObject': [
								[{'k': '$_id.unit', 'v': '$data'}]
							]
						}
					}
				}
			},
			{
				'$group': {
					'_id': '$_id.stock',
					'data': {'$mergeObjects': '$data'}
				}
			},
			# flatten document
			{'$set': {'data._id': '$_id'}},
			{
				'$replaceRoot': {
					'newRoot': '$data'
				}
			}
		]

def _all_units(filters: T2FilterModel | AllOf[T2FilterModel] | AnyOf[T2FilterModel]) -> list[str]:
	"""
	Get the set of all units involved in the selection
	"""
	if isinstance(filters, T2FilterModel):
		return [filters.unit]
	if isinstance(filters, AllOf):
		return list(set(f.unit for f in filters.all_of))
	if isinstance(filters, AnyOf):
		# NB: AnyOf may contain AllOf
		return list(set(functools.reduce(operator.iadd, (_all_units(f) for f in filters.any_of), [])))
	raise TypeError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/DisplayOptions.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                23.04.2022
# Last Modified Date:  23.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelBaseModel import AmpelBaseModel


class DisplayOptions(AmpelBaseModel):

	verbose: bool = False
	debug: bool = False
	hide_stderr: bool = False
	hide_module_not_found_errors: bool = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/DBCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.03.2021
# Last Modified Date:  20.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from argparse import ArgumentParser
from collections.abc import Sequence
from typing import Any

from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.main import AmpelArgumentParser
from ampel.core.AmpelContext import AmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag

# Help parameter descriptions
hlp = {
	'import': 'Import saved ampel databases',
	'save': 'Save a databases into file',
	'delete': 'Deletes all database starting with the provided prefix (ex: AmpelTest)',
	'prefix': 'Prefix of the collections to delete (ex: AmpelTest)',
	'config': 'Path to an ampel config file (yaml/json)',
	'secrets': 'Path to a YAML secrets store in sops format',
	'log-profile': 'One of: default, compact, headerless, verbose, debug',
	'debug': 'debug',
	'force': 'Delete potententially existing view before view creation',
	'view': 'Create or discard collection views',
	'index': 'Create or recreate collection indexes',
}

class DBCommand(AbsCoreCommand):


	@staticmethod
	def get_sub_ops() -> list[str]:
		return ['import', 'export', 'delete', 'view', 'index']


	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = self.get_sub_ops()
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'db', sub_ops, hlp, description = 'Import, export or delete ampel databases. Create or remove views.'
			)

		builder = ArgParserBuilder('db')
		builder.add_parsers(sub_ops, hlp)
		builder.notation_add_note_references()
		builder.notation_add_example_references()

		builder.req('config')
		builder.req('in', 'import')
		builder.req('out', 'export')

		builder.add_group('view', 'View arguments', sub_ops='view')
		builder.xargs(
			group='required', sub_ops='view', xargs = [
				{'name': 'create', 'action': 'store_true'},
				{'name': 'discard', 'action': 'store_true'}
			]
		)
		builder.xargs(
			group='required', sub_ops='view', xargs = [
				{'name': 'channel'},
				{'name': 'channels-or', 'nargs': '+'},
				{'name': 'channels-and', 'nargs': '+'}
			]
		)

		# Optional
		builder.opt('secrets')
		builder.opt('one-db', default=False, action='store_true')
		builder.opt('debug', action='store_true')
		builder.opt('force', 'view|index', action='store_true')
		builder.opt('col', 'index', action='append')

		builder.example('import', '-in /path/to/file')
		builder.example('export', '-out /path/to/file')
		builder.example('delete', '-mongo.prefix AmpelTest')
		builder.example('view', '-create -channel CHAN1')
		builder.example('view', '-create -channels-or CHAN1 CHAN2')
		builder.example('view', '-discard -channel CHAN1')

		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		ctx: AmpelContext = self.get_context(args, unknown_args, one_db=args['one_db'], require_existing_db=sub_op != 'index')
		db = ctx.db

		logger = AmpelLogger.from_profile(
			ctx, 'console_debug' if args['debug'] else 'console_info',
			base_flag = LogFlag.MANUAL_RUN
		)

		if sub_op == 'delete':

			logger.info(f'Deleting databases with prefix {ctx.db.prefix}')
			ctx.db.drop_all_databases()
			logger.info('Done')

		elif sub_op == 'index':

			for col_name, col_model in ctx.db.col_config.items():
				if args['col'] and col_name not in args['col']:
					continue
				col = ctx.db.get_collection(col_name)
				ctx.db.set_col_index(col, col_model, logger, force_overwrite=args['force'])

		elif sub_op == 'view':

			if 'create' not in args and 'discard' not in args:
				logger.error('Either provide "create" or "discard" in combination with view command')
				return

			try:
				if x := args.get('channel'):
					logger.info(f'{"Creating" if args["create"] else "Removing"} view for channel {x}')
					db.create_one_view(x, logger, args['force']) if args['create'] else db.delete_one_view(x, logger)
				elif x := args.get('channels_or'):
					logger.info(f'{"Creating" if args["create"] else "Removing"} view for channels {x}')
					db.create_or_view(x, logger, args['force']) if args['create'] else db.delete_or_view(x, logger)
				elif x := args.get('channels_and'):
					logger.info(f'{"Creating" if args["create"] else "Removing"} view for channels {x}')
					db.create_and_view(x, logger, args['force']) if args['create'] else db.delete_and_view(x, logger)
				else:
					logger.error('Channel(s) required\n')
					return
			except Exception as e:
				print(e)
				if 'already exists' in str(e):
					#logger.info('View already exist')
					logger.info(str(e))
				else:
					raise e

		elif sub_op in {'export', 'import'}:
			raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/AmpelContext.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.02.2020
# Last Modified Date:  11.11.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import uuid
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any
from typing_extensions import Self

from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.config.AmpelConfig import AmpelConfig
from ampel.secret.AmpelVault import AmpelVault

# Avoid cyclic import issues
if TYPE_CHECKING:
	from ampel.core.AmpelDB import AmpelDB
	from ampel.core.UnitLoader import UnitLoader


class AmpelContext:
	"""
	Execution context for Ampel processes.

	The :class:`AmpelContext` bundles together configuration, database
	connections, unit loading, and shared resources. It is typically
	created by Ampel and passed to processor units so they can access
	system-wide services.

	This class also manages runtime aliases, configuration IDs, and
	provides utilities for generating unique run identifiers. Future
	versions may support multiple :class:`~ampel.config.AmpelConfig.AmpelConfig`
	and :class:`~ampel.core.AmpelDB.AmpelDB` instances.

	.. note::
	   Auxiliary units are auto-registered globally via ``AuxUnitRegister``.
	   Run-time aliases must begin with ``%%``.
	   Convenience constructors (:meth:`load`, :meth:`build`) are provided
	   to create contexts from configuration files or distribution builders.
	"""

	def __init__(self,
		config: AmpelConfig,
		db: 'AmpelDB',
		loader: 'UnitLoader', # forward reference to avoid cyclic import issues
		resource: dict[str, Any] | None = None,
		admin_msg: str | None = None
	) -> None:
		"""
		Initialize a new context with configuration, database, and unit loader.

		:param config: ampel configuration
		:param db: ampel database object instance
		:param loader: Unit loader for instantiating units
		:param resource: Optional shared resources dictionary
		"""

		self.config = config
		self.db = db
		self.loader = loader
		self.admin_msg = admin_msg
		self.resource = resource
		self.run_time_aliases: dict[str, Any] = {}
		self.uuid = uuid.uuid4()
		
		# try to register aux units globally
		try:
			AuxUnitRegister.initialize(config)
		except Exception:
			print("UnitLoader auxiliary units auto-registration failed")  # noqa: T201


	@classmethod
	def load(cls,
		config: str | dict,
		pwd_file_path: str | None = None,
		pwds: Iterable[str] | None = None,
		freeze_config: bool = True,
		vault: AmpelVault | None = None,
		one_db: bool = False,
		**kwargs
	) -> Self:
		"""
		Instantiate a new context from a configuration file or dict.

		:param config: local path to an ampel config file (yaml or json) or loaded config as dict
		:param pwd_file_path: path to a text file containing one password per line.
		The underlying AmpelVault will be initialized with an AESecretProvider configured with these pwds.
		:param pwds: Same as 'pwd_file_path' except a list of passwords is provided directly via this parameter.
		:param freeze_config:
			whether to convert the elements contained of ampel config
			into immutable structures (:class:`dict` ->
			:class:`~ampel.view.ReadOnlyDict.ReadOnlyDict`, :class:`list` -> :class:`tuple`).
			Parameter does only apply if the config is loaded by this method, i.e if parameter 'config' is a str.
		"""

		# Avoid cyclic import issues
		from ampel.core.AmpelDB import AmpelDB  # noqa: PLC0415
		from ampel.core.UnitLoader import UnitLoader  # noqa: PLC0415

		alconf = AmpelConfig(config) if isinstance(config, dict) else AmpelConfig.load(config)
		if vault is None:
			vault = AmpelVault([])

		if pwd_file_path and not pwds:
			with open(pwd_file_path) as f:
				pwds = [l.strip() for l in f.readlines()]

		if pwds:
			# AESecretProvider is optional
			from ampel.secret.AESecretProvider import AESecretProvider  # noqa: PLC0415
			vault.providers.append(
				AESecretProvider(pwds)
			)

		db = AmpelDB.new(alconf, vault, one_db=one_db)

		return cls(
			config = alconf,
			db = db,
			loader = UnitLoader(config=alconf, db=db, vault=vault),
			**kwargs
		)


	@classmethod
	def build(cls,
		ignore_errors: bool = False,
		pwd_file_path: None | str = None,
		pwds: None | Iterable[str] = None,
		freeze_config: bool = True,
		verbose: bool = False
	) -> Self:
		"""
		Instantiate a new context by building configuration from local distributions.

		Gathers and merges all available configurations from Ampel distributions found
		locally and morphs them into a final configuration. This is a convenience method and
		not intended for production deployments.

		:param ignore_errors: If True, continue building despite non-fatal errors.
		:param pwd_file_path: Path to a text file containing one password per line.
		:param pwds: Passwords provided directly; alternative to ``pwd_file_path``.
		:param freeze_config: Whether to convert the built config into immutable structures.
		:param verbose: If True, show verbose distribution loading output.
		:returns: A new :class:`AmpelContext` instance.
		"""

		if pwd_file_path:
			with open(pwd_file_path) as f:
				pwds = [l.strip() for l in f.readlines()]

		# Import here to avoid cyclic import error
		from ampel.config.builder.DistConfigBuilder import DistConfigBuilder # noqa: PLC0415
		from ampel.config.builder.DisplayOptions import DisplayOptions # noqa: PLC0415
		cb = DistConfigBuilder(options=DisplayOptions(verbose=verbose))
		cb.load_distributions()

		return cls.load(
			cb.build_config(ignore_errors, pwds=pwds),
			pwd_file_path = pwd_file_path,
			pwds = pwds,
			freeze_config = freeze_config,
		)


	def new_run_id(self) -> int:
		"""
		Return an identifier that can be used to associate log entries from a
		single process invocation. This ID is unique and monotonicaly increasing.
		"""
		return self.db \
			.get_collection('counter') \
			.find_one_and_update(
				{'_id': 'current_run_id'},
				{'$inc': {'value': 1}},
				new=True, upsert=True
			) \
			.get('value')


	def get_config(self) -> AmpelConfig:
		"""
		.. note:: Future versions may support multiple configurations.
		"""
		return self.config


	def get_database(self) -> 'AmpelDB':
		"""
		.. note:: Future versions may support multiple databases.
		"""
		return self.db


	def resolve_conf_id(self, conf_id: int) -> dict[str, Any] | None:
		try:
			return self.config.get_conf_by_id(conf_id)
		# confid not found (obsolete or dynamically generated by process)
		except Exception:
			l = list(self.db.col_conf_ids.find({"_id": conf_id}))
			if len(l) == 0:
				return None
			del l[0]['_id']
			return l[0]


	def add_conf_id(self, conf_id: int, unit_config: dict[str, Any]) -> None:
		self.db.add_conf_id(conf_id, unit_config)
		dict.__setitem__(self.config._config["confid"], conf_id, unit_config)  # noqa: SLF001


	def add_run_time_alias(self, key: str, value: Any, overwrite: bool = False) -> None:
		if not isinstance(key, str) or key[0] != '%' != key[1]:
			raise ValueError('Run time aliases must begin with %%')
		if key in self.run_time_aliases and not overwrite:
			raise ValueError(f"Run time alias {key} already defined, set overwrite=True to ignore")
		self.run_time_aliases[key] = value


	def __repr__(self) -> str:
		return "<AmpelContext>"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/config.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.04.2023
# Last Modified Date:  06.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.core.UnitLoader import UnitLoader
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.UnitModel import UnitModel
from ampel.util.hash import build_unsafe_dict_id


def get_unit_confid(loader: 'UnitLoader', unit: str, config: dict[str, Any]) -> int:
	return build_unsafe_dict_id(
		loader.new_logical_unit(  # noqa: SLF001
			model = UnitModel(unit=unit, config=config),
			logger = AmpelLogger.get_logger()
		)._get_trace_content(),
		ret = int
	)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/demo/DemoT3Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.06.2020
# Last Modified Date:  03.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator

from ampel.abstract.AbsT3Unit import AbsT3Unit
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.struct.T3Store import T3Store
from ampel.struct.UnitResult import UnitResult
from ampel.types import T3Send, UBson
from ampel.view.SnapView import SnapView


class DemoT3Unit(AbsT3Unit[SnapView]):

	parameter: int = 10

	def process(self,
		gen: Generator[SnapView, T3Send, None],
		t3s: T3Store
	) -> UBson | UnitResult:

		self.logger.info(f"DemoT3Unit output (parameter={self.parameter}):")

		for v in gen:

			self.logger.info("id: " + str(v.id))
			gen.send(
				# This journal customization will be applied only to the current 'stock'
				JournalAttributes(tag="DemoT3SpecificTag", extra={'a': 1})
			)

		# This journal customization will be applied to all stocks
		return UnitResult(
			body = {'param': 'value'},
			code = 10,
			tag = "T3DocTag",
			journal = JournalAttributes(tag="DemoT3UnitTag")
		)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/debug.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                24.07.2022
# Last Modified Date:  08.06.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>


import os
import sys
import pdb
import traceback
import multiprocessing.pool
from cProfile import Profile
from pstats import Stats

from ampel.cli.utils import _maybe_int


"""
Usage:
from ampel.util.debug import mp_breakpoint
-> mp_breakpoint()

from ampel.util.debug import trace_prints
-> trace_prints()
"""


class MockPool(multiprocessing.pool.Pool):

	def __init__(self, **kwargs):
		self._terminate = self._check_running = lambda: True
		if 'initializer' in kwargs:
			kwargs['initializer'](*kwargs['initargs'])

	def apply_async(self, func, args=(), kwds=None):
		return MockFuture(func, args, kwds)


class MockFuture:

	def __init__(self, func, args=(), kwds=None):
		self.func = func
		self.args = args
		self.kwds = kwds

	def get(self, **kwargs):
		if isinstance(self.kwds, dict):
			return self.func(**self.kwds)
		return self.func(*self.args)


# https://stackoverflow.com/questions/26289153/how-to-use-ipdb-set-trace-in-a-forked-process
class ForkedPdb(pdb.Pdb):
	"""
	A Pdb subclass that may be used within a multiprocessing process

	Forword: If you do not use multiprocessing, just use breakpoint()

	Usage:
	insert: "ForkedPdb().set_trace()" at the wished position
	<run your job>
	a pdb shell should pop up, type "interact" and press enter
	<do debug>
	when you are done, press CTRL-D
	you are back in the pdb shell, type "continue" and press enter
	"""

	def interaction(self, *args, **kwargs):
		_stdin = sys.stdin
		try:
			sys.stdin = open('/dev/stdin') # noqa: SIM115
			print(f"ForkedPdb started (pid: {os.getpid()})")  # noqa: T201
			pdb.Pdb.interaction(self, *args, **kwargs)
		finally:
			sys.stdin = _stdin


# https://stackoverflow.com/questions/1617494/finding-a-print-statement-in-python
class TracePrints:

	def __init__(self):
		self.stdout = sys.stdout

	def write(self, s):
		self.stdout.write(f"Writing {s!r}\n")
		traceback.print_stack(file=self.stdout)


def trace_prints():
	original_stdout.flush()
	sys.stdout = TracePrints()

def stop_trace_prints():
	original_stdout.flush()
	sys.stdout = original_stdout


def start_profiling() -> Profile:
	cprofile = Profile()
	cprofile.enable()
	return cprofile


def report_stats(
	cprofile: Profile,
	option: str | int | bool,
	suffix: str | None = None
) -> None:

	cprofile.disable()
	if isinstance(_maybe_int(option), int):
		stats = Stats(cprofile)
		stats.sort_stats('tottime').print_stats(option)
	elif isinstance(option, str):
		cprofile.dump_stats(f'{option}{suffix}.cprof')
	elif option is True:
		cprofile.print_stats('tottime')


mp_breakpoint = ForkedPdb().set_trace
original_stdout = sys.stdout



#!/usr/bin/env python

from typing import Generic, Protocol, TypeVar

_T_contra = TypeVar("_T_contra", contravariant=True)

class DocIngesterProtocol(Protocol, Generic[_T_contra]):

	def ingest(self, doc: _T_contra) -> None:
		...




import copy

import pytest

from ampel.config.alter.HashT2Config import HashT2Config
from ampel.config.AmpelConfig import AmpelConfig
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.test.DummyStateT2Unit import DummyStateT2Unit


def test_load_old_configids(mock_context: DevAmpelContext, ampel_logger):
    """
    Configuration hashes stored via HashT2Config.alter are loaded from the
    database when a new DevAmpelContext is instantiated
    """
    unit_config = {"foo": 37}
    mock_context.register_unit(DummyStateT2Unit)
    pre_register_context = DevAmpelContext(
        config=AmpelConfig(copy.deepcopy(mock_context.config.get())),
        db=mock_context.db,
        loader=mock_context.loader,
    )

    altered = HashT2Config().alter(
        mock_context,
        dict(
            directives=[
                dict(
                    channel="TEST",
                    ingest=dict(
                        combine=[
                            dict(
                                unit="T1SimpleCombiner",
                                state_t2=[dict(unit="DummyStateT2Unit", config=unit_config)],
                            )
                        ]
                    )
                )
            ]
        ),
        ampel_logger
    )

    hashed_unit_config = altered['directives'][0]['ingest']['combine'][0]['state_t2'][0]['config']
    assert isinstance(hashed_unit_config, int)

    post_register_context = DevAmpelContext(
        config=AmpelConfig(copy.deepcopy(pre_register_context.config.get())),
        db=mock_context.db,
        loader=mock_context.loader,
    )
    # config was not present before registration
    with pytest.raises(ValueError, match=r"Config with id .* not found"):
        pre_register_context.config.get_conf_by_id(hashed_unit_config)

    # config is present after registration in the context where it was registered
    unit_config_from_confid = mock_context.config.get_conf_by_id(hashed_unit_config)
    assert unit_config_from_confid == unit_config

    # also in an unrelated context instantiated after registration
    assert (
        post_register_context.config.get_conf_by_id(hashed_unit_config) == unit_config
    ), "unregistered configs loaded from database"



from collections.abc import Callable

from ampel.queue.AbsProducer import AbsProducer


class NullProducer(AbsProducer):
    """
    A producer that does nothing.
    """

    def produce(
        self, item: AbsProducer.Item, delivery_callback: None | Callable[[], None]
    ) -> None:
        if delivery_callback:
            delivery_callback()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return None


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/time/UnixTimeModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.09.2018
# Last Modified Date:  06.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Literal

from ampel.base.AmpelBaseModel import AmpelBaseModel


class UnixTimeModel(AmpelBaseModel):

	match_type: Literal['unix_time']
	value: int

	def get_timestamp(self, **kwargs) -> int:
		return self.value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/demo/DemoProcessor.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.03.2021
# Last Modified Date:  01.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import time
from random import randint
from typing import Any

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsT3Unit import AbsT3Unit
from ampel.core.EventHandler import EventHandler
from ampel.log import AmpelLogger
from ampel.model.UnitModel import UnitModel


class DemoProcessor(AbsEventUnit):

	parameter_a: int
	parameter_b: int = 200

	def proceed(self, event_hdlr: EventHandler) -> Any:

		self.parameter_b = randint(3, 10)
		time.sleep(self.parameter_b)

		# Processor/Admin unit must instantiate their own loggger
		logger = AmpelLogger.get_logger()

		# Feedback
		logger.info("Executing run()")

		# Processor units have access to the associated process id
		logger.info(f"Requested by process {self.process_name}")

		# Config parameters
		logger.info(f"Mandatory config parameter 'parameter_a' value: {self.parameter_a}")
		logger.info(f"Optional config parameter 'parameter_b' value: {self.parameter_b}")


		# Process/Admin units have access to AmpelContext
		ctx = self.context

		# Which contains an instance of UnitLoader
		loader = ctx.loader

		# With which base units can be instantiated
		unit = loader.new_logical_unit(
			model = UnitModel(unit = "DemoT3Unit"),
			logger = logger,
			sub_type = AbsT3Unit
		)

		print(unit)  # noqa: T201



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/handlers/EnclosedChanRecordBufHandler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.05.2020
# Last Modified Date:  15.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from logging import Handler, Logger

from ampel.log.handlers.RecordBufferingHandler import RecordBufferingHandler
from ampel.protocol.LoggingHandlerProtocol import LoggingHandlerProtocol
from ampel.types import ChannelId, StockId


class EnclosedChanRecordBufHandler(RecordBufferingHandler):

	__slots__ = '_channel', '_unit', '_empty_msg'


	def __init__(self, level: int, channel: ChannelId, unit: None | str = None) -> None:
		super().__init__(level)
		self._channel = channel
		self._unit = unit
		self._empty_msg = {'c': channel}


	def forward(self,
		target: Logger | Handler | LoggingHandlerProtocol,
		stock: None | StockId = None,
		extra: None | dict = None,
		clear: bool = True
	) -> None:
		"""
		Forwards saved log records to provided logger/handler instance.
		Clears the internal record buffer.
		"""
		for rec in self.buffer:

			if rec.levelno >= target.level:

				if rec.msg:
					rec.msg = {'c': self._channel, 't': rec.msg}
				else:
					rec.msg = self._empty_msg

				rec.stock = stock

				if extra:
					if 'extra' in rec.__dict__:
						rec.extra = {**rec.extra, **extra} # type: ignore[union-attr,dict-item]
					else:
						rec.extra = extra

				if self._unit:
					rec.unit = self._unit

				target.handle(rec) # type: ignore[arg-type]

		if clear:
			self.buffer.clear()
			self.has_error = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/ingest/T2Compiler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                01.01.2018
# Last Modified Date:  21.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import datetime
from typing import Any, NamedTuple

from ampel.abstract.AbsCompiler import AbsCompiler, ActivityRegister
from ampel.content.MetaActivity import MetaActivity
from ampel.content.T2Document import T2Document
from ampel.enum.DocumentCode import DocumentCode
from ampel.protocol.DocIngesterProtocol import DocIngesterProtocol
from ampel.types import ChannelId, StockId, T2Link, UnitId


class T2Compiler(AbsCompiler):
	"""
	Helps build a minimal set of :class:`T2Document <ampel.content.T2Document.T2Document>`
	Multiple T2 documents associated with the same stock record accross different channels
	are merged into one single T2 document that references all corresponding channels.
	"""

	class UnitKey(NamedTuple):
		unit: UnitId
		config: None | int
		link: T2Link
		stock: StockId
	
	class DocInfo(NamedTuple):
		channels: set[ChannelId]
		ttl: None | datetime.timedelta
		meta: dict[
			frozenset[tuple[str, Any]], # key: traceid
			tuple[
				ActivityRegister,
				dict[str, Any]
			]
		]		

	col: None | str = None

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.t2s: dict[T2Compiler.UnitKey, T2Compiler.DocInfo] = {}

	def add(self, # type: ignore[override]
		unit: UnitId,
		config: None | int,
		stock: StockId,
		link: T2Link,
		channel: ChannelId,
		ttl: None | datetime.timedelta,
		traceid: dict[str, Any],
		activity: None | MetaActivity | list[MetaActivity] = None,
		meta_extra: None | dict[str, Any] = None
	) -> None:
		"""
		:param tag: tag(s) to be added to T2Document. A corresponding dedicated channel-less
		meta entry will be created (with actin MetaActionCode.ADD_INGEST_TAG and/or ADD_OTHER_TAG)
		"""

		# Doc id
		k = T2Compiler.UnitKey(unit, config, link, stock)
		tid = frozenset(traceid.items())

		# a: tuple({channels}, dict[traceid, (activity register, meta_extra)])
		if (a := self.t2s.get(k)) is not None:

			a.channels.add(channel) # set of all channels

			# One meta record will be created for each unique trace id
			# (the present compiler could be used by multiple handlers configured differently)
			if tid in a.meta:

				# (activity register, meta_extra)
				b = a.meta[tid]

				# update internal register
				self.register_meta_info(b[0], b[1], channel, activity, meta_extra)

			else:
				a.meta[tid] = self.new_meta_info(channel, activity, meta_extra)

			if ttl is not None and (a.ttl is None or ttl > a.ttl):
				self.t2s[k] = T2Compiler.DocInfo(a.channels, ttl, a.meta)
		else:
			self.t2s[k] = T2Compiler.DocInfo({channel}, ttl, {tid: self.new_meta_info(channel, activity, meta_extra)})


	def commit(self, ingester: DocIngesterProtocol[T2Document], now: int | float, **kwargs) -> None:

		for k, v in self.t2s.items():

			meta, tags = self.build_meta(v.meta, now)

			# Note: mongodb maintains key order
			d: T2Document = {
				'stock': k.stock,
				'unit': k.unit,
				'config': k.config,
				'link': k.link,
				'channel': list(v.channels),
				'meta': meta,
				'code': DocumentCode.NEW,
				'body': [],
			}

			if self.col:
				d['col'] = self.col

			if self.origin:
				d['origin'] = self.origin

			if v.ttl is not None:
				d['expiry'] = datetime.datetime.fromtimestamp(
					now, tz=datetime.timezone.utc
				) + v.ttl

			if tags:
				d['tag'] = tags

			ingester.ingest(d)

		self.t2s.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/handlers/DefaultRecordBufferingHandler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                05.05.2020
# Last Modified Date:  15.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.log.handlers.RecordBufferingHandler import RecordBufferingHandler
from ampel.protocol.LoggingHandlerProtocol import LoggingHandlerProtocol
from ampel.types import ChannelId, StockId


class DefaultRecordBufferingHandler(RecordBufferingHandler):
	"""
	MemoryHandler-like class that can grown infinitely and features a convenience method called 'forward'
	"""

	__slots__ = '_extra', '_unit'


	def __init__(self, level: int, extra: None | dict[str, Any] = None, unit: None | str = None) -> None:
		super().__init__(level)
		self._extra = extra or {}
		self._unit = unit


	def forward(self,
		target: LoggingHandlerProtocol,
		channel: None | ChannelId = None,
		stock: None | StockId = None,
		extra: None | dict = None,
		clear: bool = True
	) -> None:
		"""
		Forwards saved log records to provided logger/handler instance.
		Clears the internal record buffer.
		"""
		for rec in self.buffer:

			if rec.levelno < target.level:
				continue

			if channel:
				rec.channel = channel

			if stock:
				rec.stock = stock

			if extra:
				if hasattr(rec, 'extra') and rec.extra:
					rec.extra |= extra
				else:
					rec.extra = extra

			if self._extra:
				if hasattr(rec, 'extra') and rec.extra:
					rec.extra |= self._extra
				else:
					rec.extra = self._extra

			if self._unit:
				rec.unit = self._unit

			target.handle(rec)

		if clear:
			self.buffer.clear()
			self.has_error = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/aux/filter/AbsLogicOperatorFilter.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.01.2020
# Last Modified Date:  18.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from typing import Any, Generic

from ampel.abstract.AbsApplicable import AbsApplicable
from ampel.base.decorator import abstractmethod
from ampel.model.aux.FilterCriterion import FilterCriterion
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.FlatAnyOf import FlatAnyOf
from ampel.types import T


# mypy: disable-error-code = empty-body
class AbsLogicOperatorFilter(AbsApplicable, Generic[T], abstract=True):

	filters: FilterCriterion | FlatAnyOf[FilterCriterion] | AllOf[FilterCriterion]

	@staticmethod
	@abstractmethod
	def _apply_filter(args: Sequence[T], f: FilterCriterion) -> list[T]:
		...

	def apply(self, args: Sequence[T]) -> list[T]:

		if isinstance(self.filters, FilterCriterion):
			return self._apply_filter(args, self.filters)

		if isinstance(self.filters, AllOf):
			ret: Any = args
			for f in self.filters.all_of:
				ret = self._apply_filter(ret, f)
			return ret

		if isinstance(self.filters, FlatAnyOf):
			ret = []
			for f in self.filters.any_of:
				for el in self._apply_filter(args, f):
					if el not in ret:
						ret.append(el)
			return ret

		raise ValueError("Incorrect type provided for parameter 'filters'")



import pytest

from ampel.core.AmpelContext import AmpelContext
from ampel.model.UnitModel import UnitModel


@pytest.mark.usefixtures("_patch_mongo")
def test_validate(core_config):
    ctx = AmpelContext.load(core_config)
    with ctx.loader.validate_unit_models():
        UnitModel(unit="T3LatestStateDataLoader", config={"directives": []})



from collections.abc import Iterable
from threading import Event
from typing import Generic

from ampel.abstract.AbsContextManager import AbsContextManager
from ampel.base.AmpelUnit import AmpelUnit
from ampel.base.decorator import abstractmethod
from ampel.types import T, Traceless


class AbsConsumer(AbsContextManager, AmpelUnit, Generic[T], abstract=True):

	#: Event that can be set to break out of consume()
	stop: Traceless[Event]

	@abstractmethod
	def consume(self) -> None | T:
		"""Get a single message from the queue, returning None if the queue is empty, or stop is set"""
		...
	
	@abstractmethod
	def acknowledge(self, docs: Iterable[T]) -> None:
		"""Acknowledge the processing of a batch of messages"""
		...


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/utils/stock.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.04.2020
# Last Modified Date:  29.04.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from multiprocessing import Pool

from pymongo import MongoClient
from pymongo.collection import Collection

from ampel.types import ChannelId, OneOrMany, StockId


def get_ids_using_find(
	col: Collection,
	channel: OneOrMany[ChannelId],
	batch_size=1000000
) -> dict[ChannelId, set[StockId]]:

	if isinstance(channel, int | str):
		return {
			channel: {
				el['_id'] for el in col \
					.find({'channel': channel}, {'_id': 1}) \
					.batch_size(batch_size)
			}
		}

	return {k: get_ids_using_find(col, k, batch_size)[k] for k in channel}


def get_ids_using_parallel_find(
	channel: OneOrMany[ChannelId],
	mongo_uri: None | str = None,
	db_name: str = 'Ampel_data',
	col_name: str = 'stock', *,
	pool_size: None | int = None,
	batch_size: int = 1000000
) -> dict[ChannelId, set[StockId]]:

	if isinstance(channel, int | str):
		return get_ids_using_find(
			MongoClient(mongo_uri) \
				.get_database(db_name) \
				.get_collection(col_name),
			channel, batch_size
		)

	pool = Pool(processes=pool_size)

	ret: dict[ChannelId, set[StockId]] = {k: set() for k in channel}

	results = [
		pool.apply_async(
			find_ids_worker,
			(k, mongo_uri, db_name, col_name, batch_size),
			callback = ret[k].update
		)
		for k in channel
	]

	for r in results:
		r.wait()

	pool.close()
	return ret


def find_ids_worker(
	channel: OneOrMany[ChannelId],
	mongo_uri: None | str = None,
	db_name: str = 'Ampel_data',
	col_name: str = 'stock',
	batch_size: int = 1000000
) -> set[StockId]:

	return { # type: ignore[var-annotated]
		el['_id']
		for el in MongoClient(mongo_uri)
			.get_database(db_name)
			.get_collection(col_name)
			.find({'channel': channel}, {'_id': 1})
			.batch_size(batch_size)
	}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/MongoOpts.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  13.08.2022
# Last Modified By:    jvs

from ampel.base.AmpelBaseModel import AmpelBaseModel


class MongoOpts(AmpelBaseModel):
    reset: bool = False
    prefix: str = "Ampel"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/BufferCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2021
# Last Modified Date:  03.11.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys
from argparse import ArgumentParser
from collections.abc import Sequence
from typing import Any

from ampel.cli.AbsLoadCommand import AbsLoadCommand
from ampel.cli.AbsStockCommand import AbsStockCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.utils import maybe_load_idmapper
from ampel.core.AmpelContext import AmpelContext
from ampel.log.LogFlag import LogFlag
from ampel.model.UnitModel import UnitModel
from ampel.t3.T3Processor import T3Processor

hlp = {
	'config': 'Path to an ampel config file (yaml/json)',
	# Optional
	'secrets': 'Path to a YAML secrets store in sops format',
	'db': 'Database prefix. If set, "-mongo.prefix" value will be ignored',
	'out': 'Path to file where serialized views will be saved (printed to stdout otherwise)',
	'log-profile': 'One of: default, compact, headerless, verbose, debug',
	'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
	'binary': 'Store buffers as compact BSON structures',
	'pretty': 'Show buffers using spacier prettyjson',
	'getch': 'Wait for a key to be pressed in terminal before printing next buffer (not compatible with binary option)'
}

class BufferCommand(AbsStockCommand, AbsLoadCommand):


	@staticmethod
	def get_sub_ops() -> list[str]:
		return ['show', 'save']


	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = self.get_sub_ops()
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'buffer', sub_ops, hlp, description = 'Show or save ampel buffers.'
			)

		builder = ArgParserBuilder('buffer')
		hlp.update(self.get_select_args_help())
		hlp.update(self.get_load_args_help())
		builder.add_parsers(sub_ops, hlp)

		builder.req('config')
		builder.req('out', 'save', default=True)

		builder.opt('secrets', default=None)
		builder.opt('id-mapper')
		builder.opt('debug', action='store_true')
		builder.opt('log-profile', default='default')
		builder.opt('binary', 'save', action='store_true')
		builder.opt('pretty', 'show', action='store_true')
		builder.opt('getch', 'show', action='store_true')

		# Selection args
		self.add_selection_args(builder)

		# Content args
		self.add_load_args(builder, 'Views content arguments')

		# Example
		builder.example('show', '-stock 85628462 -no-t0')
		builder.example('show', '-db MyDB -stock 519059889 -no-t0')
		builder.example('save', '-stock 85628462 -no-t0 -out ZTFabcdef.json')
		
		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		ctx = self.get_context(
			args, unknown_args, ContextClass = AmpelContext,
			require_existing_db = args['db'] or True, one_db='auto'
		)

		maybe_load_idmapper(args)

		conf = {
			'fd': open(args['out'], 'wb' if args.get('binary') else 'w') \
				if sub_op == 'save' else sys.stdout, # noqa: SIM115
			'id_mapper': args['id_mapper'],
			'verbose': sub_op == 'save',
			'binary': args.get('binary') or False
		}

		if args.get('pretty'):
			conf['pretty'] = True

		if not args.get('binary') and args.get('getch'):
			conf['getch'] = True

		t3p = T3Processor(
			context = ctx,
			process_name = 'BufferCommand',
			template = 'compact_t3',
			base_log_flag = LogFlag.MANUAL_RUN,
			log_profile = 'console_debug' if args.get('debug') else 'console_info',
			supply = UnitModel(
				unit = 'T3DefaultBufferSupplier',
				config = {
					'select': self.build_select_model(args),
					'load': self.build_load_model(args)
				}
			),
			stage = UnitModel(
				unit = 'T3BufferExporterStager',
				config = conf
			)
		)

		t3p.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/AbsLoadCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2021
# Last Modified Date:  20.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.model.UnitModel import UnitModel

# Maps args with load aliases
arg_doc_mapping = {
	'no_stock': "STOCK",
	'no_t0': "DATAPOINT",
	'no_t1': "COMPOUND",
	'no_t2': "T2DOC"
}

class AbsLoadCommand(AbsCoreCommand, abstract=True):
	"""
	Base for commands that load ampel data using AbsT3Loader subclasses
	"""

	@staticmethod
	def get_load_args_help() -> dict[str, str]:
		return {
			'latest': 'Include only latest state (*)',
			'no-stock': 'Exclude stock document from view',
			'no-plots': 'Exclude plots',
			'no-logs': 'Exclude logs'
		}

	def add_load_args(self,
		builder: ArgParserBuilder,
		group_description: str,
		sub_ops: str = 'all'
	) -> None:

		# Content args
		builder.add_group('content', group_description, sub_ops=sub_ops)
		builder.arg('latest', group='content', sub_ops=sub_ops, action='store_true')
		builder.arg('no-stock', group='content', sub_ops=sub_ops, action='store_true')
		for el in (0, 1, 2):
			builder.arg(
				f'no-t{el}', group='content', sub_ops=sub_ops,
				action='store_true', help=f"Exclude t{el} documents from view"
			)
		builder.arg('no-plots', group='content', sub_ops=sub_ops, action='store_true')
		builder.arg('no-logs', group='content', sub_ops=sub_ops, action='store_true', help="Exclude logs")

		# Note
		builder.note("Latest state means... [adequate description in a few words]")


	def build_load_model(self, args: dict[str, Any]) -> UnitModel:
		return UnitModel(
			unit = "T3LatestStateDataLoader" if args.get("latest") else "T3SimpleDataLoader",
			config = {
				'channel': args['channel'],
				'directives': [
					v for k, v in arg_doc_mapping.items() if not args[k]
				]
			}
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/distrib.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.03.2021
# Last Modified Date:  14.11.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import os, re, importlib, json
from typing import TypeAlias
from collections.abc import Generator, MutableSequence, Sequence, Iterable
from importlib import metadata
from pathlib import Path

# NB: PackagePath implements read_text(), but is not a subclass of Path
PathLike: TypeAlias = Path | metadata.PackagePath
PathList: TypeAlias = MutableSequence[PathLike]


def get_dist_names(distrib_prefixes: Sequence[str] = ("ampel-", "pyampel-")) -> list[str]:
	""" Get all installed distributions whose names start with the provided prefix """

	# ensure that at least interface and core are found
	prefixes = {"ampel-interface", "ampel-core"}.union(distrib_prefixes)
	ret = list({ # Note: list(set) because editable installs are found twice in distributions()
		dist.name for dist in metadata.distributions()
		if any(dist.name.startswith(prefix) for prefix in prefixes)
	})

	if ret:
		ret.insert(0, ret.pop(next(i for i, el in enumerate(ret) if "core" in el)))
		ret.insert(0, ret.pop(next(i for i, el in enumerate(ret) if "interface" in el)))

	return ret


def get_files(
	dist_name: str,
	lookup_dir: str | None = None,
	pattern: re.Pattern | None = None
) -> PathList:
	""" Loads all known conf files of the provided distribution (name) """

	if lookup_dir and not lookup_dir.endswith("/"):
		lookup_dir += "/"

	files: PathList = []

	for dist in metadata.distributions():

		if dist.name != dist_name:
			continue

		# Try direct_url.json first
		mod_path = resolve_direct_url_path(dist_name)
		if mod_path and os.path.isdir(mod_path):
			files.extend(walk_dir(mod_path, lookup_dir, pattern))
			break  # skip .pth parsing if direct path works

		# dist.files may be None in 3.10
		if dist.files is None:
			# fallback: look directly in site-packages
			sp = Path(str(dist.locate_file("")))
			conf_dir = sp / (lookup_dir or "") / dist_name
			if conf_dir.exists():
				files.extend(walk_dir(str(conf_dir), lookup_dir, pattern))
			continue

		for p in dist.files or []:
			if p.suffix == ".pth":
				pth_path = Path(str(dist.locate_file(p)))
				lines = pth_path.read_text().splitlines()
				for line in lines:
					line = line.strip() # noqa PLW2901
					if os.path.isdir(line):
						files.extend(walk_dir(line, lookup_dir, pattern))
					elif line.startswith("import __editable__"):
						mod_path = resolve_editable_hook_path(pth_path)
						if not mod_path:
							mod_path = resolve_direct_url_path(dist_name)
						if mod_path:
							files.extend(walk_dir(mod_path, lookup_dir, pattern))
						else:
							raise ValueError(
								f"Could not resolve module path from editable hook or direct_url.json "
								f"in {pth_path}. Try installing {dist_name} normally (non-editable)."
							)
					else:
						raise ValueError(
							f"Invalid path '{line}' in {pth_path}. "
							f"If {dist_name} was installed with setuptools, it must be installed normally (not in editable mode)."
						)
			elif _check_match(p.as_posix(), lookup_dir, pattern):
				files.append(Path(str(dist.locate_file(p))))

	return files


def _check_match(
	path: str,
	lookup_dir: str | None,
	pattern: re.Pattern | None
) -> bool:
	if lookup_dir and not path.startswith(lookup_dir):
		return False
	return not (pattern and not pattern.match(path))


def walk_dir(
	path: str,
	lookup_dir: str | None = None,
	pattern: re.Pattern | None = None
) -> Generator[Path, None, None]:

	for root, _, files in os.walk(path):
		for fname in files:
			full_path = os.path.join(root, fname)
			rel_path = os.path.relpath(full_path, path)
			if _check_match(rel_path, lookup_dir, pattern):
				yield Path(full_path)


def resolve_editable_hook_path(pth_path: Path) -> str | None:
	try:
		text = pth_path.read_text().strip()
		match = re.search(r'import\s+(__editable__[^;]+)', text)
		if match:
			finder_module = match.group(1)
			finder = importlib.import_module(finder_module)
			if finder.__file__ is not None:
				return os.path.dirname(finder.__file__)
	except Exception:
		pass
	return None


def resolve_direct_url_path(dist_name: str) -> str | None:
	try:
		dist = metadata.distribution(dist_name)
		direct_url = dist.read_text("direct_url.json")
		if direct_url:
			info = json.loads(direct_url)
			if info.get("dir_info", {}).get("editable") and info.get("url", "").startswith("file://"):
				return info["url"][7:]  # strip 'file://'
	except Exception:
		pass
	return None


def get_classes_info_from_dist(
	pkg_name: str,
	sub_dir: str = "ampel",
	parent_map: dict[str, list[str]] | None = None
) -> dict[str, list[str]]:
	"""
	Returns a mapping from class file paths to their direct parent class names.

	This function scans a subdirectory within the specified installed package and identifies
	class definitions that follow the Ampel convention: each file defines a single class whose
	name matches the filename (excluding the `.py` extension). For each matching class,
	it extracts its direct parent class names and stores them in a dictionary.

	The result is a mapping where:
	  - keys are file paths to class definitions
	  - values are lists of direct parent class names, with generic type annotations removed

	:param pkg_name: Name of the installed package to inspect (e.g., 'ampel-interface').
	:param sub_dir: Subdirectory within the package to scan for class files.
	:param parent_map: Optional dictionary to populate with results. If provided, it will be updated.

	:return: A dictionary mapping each class file path to a list of its direct parent class names.

	:example:
		>>> get_classes_info_from_dist("ampel-interface")
		{
			'/Path/Ampel-interface/ampel/abstract/AbsT0Unit.py': ['LogicalUnit', 'abstract=True'],
			'/Path/Ampel-interface/ampel/abstract/AbsIdMapper.py': ['AmpelABC', 'AmpelBaseModel', 'abstract=True'],
			...
		}
	"""

	if parent_map is None:
		parent_map = {}

	for f in get_files(pkg_name, sub_dir, re.compile(r'.*/[A-Z][^/]*\.py$')):

		try:
			content = f.read_text()
		except Exception:
			continue  # skip unreadable files

		if match := re.search(rf'class\s+{f.stem}\((.*?)\):', content):
			raw = match.group(1)
			pattern = r'\[.*?\]' # handle AbsT3Supplier[Generator[AmpelBuffer, None, None]]
			while re.search(pattern, raw):
				raw = re.sub(pattern, '', raw)
			parents = [p.strip() for p in raw.replace(']', '').replace('[', '').split(',')]
			parent_map[str(f)] = parents

	return parent_map


def resolve_class_ancestry(
	file_path: str,
	parent_map: dict[str, list[str]],
	visited: set[str] | None = None
) -> list[str]:
	"""
	Recursively resolves the full inheritance chain for a class identified by its file path.

	This function traverses the `parent_map` to collect all direct and indirect parent class names
	for the class defined in the given file. It follows the Ampel convention: each file defines a
	single class whose name matches the filename (excluding the `.py` extension).

	:param file_path: Path to the class file, used as a key in `parent_map`.
	:param parent_map: Mapping of file paths to lists of direct parent class names.
	:param visited: Internal set used to track visited file paths and avoid cycles.

	:return: A flattened list of all inherited class names (direct and indirect).

	:example:
	>>> get_class_ancestry('/Path/Ampel-HU-cosmo/ampel/contrib/hu/t3/T3CosmicDipole.py', parent_map)
	['AbsT3Unit', 'LogicalUnit', 'AmpelABC', 'AmpelUnit', 'Generic[T]', 'SNeCosmicFlow']
	"""

	if visited is None:
		visited = set()

	if file_path in visited:
		return []

	visited.add(file_path)

	direct_parents = parent_map.get(file_path, [])
	resolved = []

	for parent in direct_parents:
		resolved.append(parent)
		# Try to find a file path in parent_map whose stem matches the parent class name
		for candidate_path in parent_map:
			if Path(candidate_path).stem == parent:
				resolved.extend(resolve_class_ancestry(candidate_path, parent_map, visited))
				break  # stop after first match

	return resolved


def get_classes_ancestry(
	parent_map: dict[str, list[str]],
	ancestry_map: dict[str, list[str]] | None = None,
	required_parents: Iterable[str] | None = None,
	exclude_parents: Iterable[str] | None = None,
	exclude_abstract: bool = True
) -> dict[str, list[str]]:
	"""
	Resolves the full inheritance chain for all classes defined in a parent map.

	This function iterates over each entry in `parent_map` and uses `resolve_class_ancestry`
	to collect all direct and indirect parent class names. It assumes the Ampel convention:
	each file defines a single class whose name matches the filename (excluding the `.py` extension).

	Filtering options:
	- If `required_parents` is provided, only classes whose ancestry includes at least one of the
	  specified parent names will be included.
	- If `exclude_parents` is provided, any class whose ancestry includes one of the excluded
	  parent names or whose name matches an excluded parent will be skipped.
	- If `exclude_abstract` is True, any class whose ancestry includes the marker "abstract=true"
	  will be excluded.

	:param parent_map: Dictionary mapping file paths to lists of direct parent class names.
	:param ancestry_map: Optional dictionary to populate with resolved inheritance chains.
	If not provided, a new one is created.
	:param required_parents: Optional list of parent class names to filter the output.
	:param exclude_parents: Optional list of parent class names to exclude from the output.
	:param exclude_abstract: If True, excludes classes marked as abstract.

	:return: A dictionary mapping each class file path to its full list of inherited class names.
	"""

	if ancestry_map is None:
		ancestry_map = {}

	# PLC0206: unpacking items() every loop less efficient than single key-based value access
	for fname in parent_map: # noqa: PLC0206

		# Ancestry was already resolved
		if fname in ancestry_map:
			continue

		# Filter by abstract marker
		if exclude_abstract and 'abstract=True' in parent_map[fname]:
			continue

		resolved = resolve_class_ancestry(fname, parent_map)

		# Filter by required parents
		if required_parents is not None and not any(f in resolved for f in required_parents):
			continue

		# Filter by excluded parents or class name
		if exclude_parents is not None:
			class_name = fname.split("/")[-1].removesuffix(".py")
			if any(f in resolved for f in exclude_parents) or class_name in exclude_parents:
				continue

		ancestry_map[fname] = resolved
		# print(fname, resolved)

	return ancestry_map



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/project/T3ChannelProjector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.01.2020
# Last Modified Date:  22.11.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

from collections.abc import Sequence
from typing import Any

from ampel.aux.ComboDictModifier import ComboDictModifier
from ampel.log import VERBOSE
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.t3.stage.project.T3BaseProjector import T3BaseProjector
from ampel.types import ChannelId
from ampel.util.logicschema import reduce_to_set


class T3ChannelProjector(T3BaseProjector):

	channel: ChannelId | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId]

	# Whether to cast structures into immutables objects after modification
	freeze: bool = True

	# Whether fields from input ampel_buffer can be directly altered/modified
	# This is the case if the parent T3UnitRunner contains only one run block
	# If this is not the case, new dict instances must be created to modify existing dicts
	unalterable: bool = True


	def __init__(self, **kwargs) -> None:
		"""
		:param channel: the channel id of the channel for which projections should be performed
		:param unalterable: see ComboDictModifier docstring
		:param freeze: see ComboDictModifier docstring
		"""

		super().__init__(**kwargs)

		self.verbose = self.logger.verbose
		if self.verbose:
			self.logger.log(VERBOSE, f"Setting up channel project for '{self.channel}'")
		self._channel_set: set[ChannelId] = reduce_to_set(self.channel)

		journal_modifier = ComboDictModifier(
			logger = self.logger,
			unalterable = self.unalterable,
			freeze = self.freeze,
			modifications = [
				# Modified ex: {"HU_RANDOM": 3213143434, "HU_RAPID": 43789574389}
				ComboDictModifier.KeepOnlyModel(op="keep_only", key="ts", keep=list(self._channel_set)),
				# Added ex: {"msg": "test", "tier":0, "channel": ["HU_RANDOM", "HU_RAPID"]}
				ComboDictModifier.FuncModifyModel(op="modify", key="journal", func=self.channel_projection),
				ComboDictModifier.FuncModifyModel(op="modify", key="channel", func=self.overwrite_root_channel)
			]
		)

		self.add_func_projector("stock", journal_modifier.apply, first=True)

		for key in ("t1", "t2"):
			self.add_func_projector(key, self.channel_projection, first=True)


	def overwrite_root_channel(self, v: Sequence[ChannelId]) -> None | Sequence[ChannelId]:
		return subset if (subset := list(self._channel_set.intersection(v))) else None


	def channel_projection(self, dicts: Sequence[dict[str, Any]]) -> Sequence[dict[str, Any]]:
		"""
		Filters out dict entries not associated with configured channel
		Note: debug ouput handled by super class
		"""

		channel_set = self._channel_set
		setitem = dict.__setitem__
		ret: list[dict] = []

		if not dicts:
			return []

		for el in dicts:
			if elchan := el.get('channel'):
				if isinstance(elchan, str | int):
					if elchan in channel_set:
						ret.append(el)
				elif subset := channel_set.intersection(elchan):
					channels = (
						list(subset)
						if len(subset) > 1 else next(iter(subset)) # type: ignore[arg-type]
					)
					if self.unalterable:
						ret.append({**el, 'channel': channels})
					else:
						setitem(el, 'channel', channels)
						ret.append(el)

		return tuple(ret)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/T3BufferExporterUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.08.2022
# Last Modified Date:  17.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator

from pydantic import TypeAdapter

from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.abstract.AbsT3Unit import AbsT3Unit
from ampel.cli.export import bin_export, get_fd, txt_export
from ampel.cli.T3BufferExporterStager import TextExportOptions
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.struct.T3Store import T3Store
from ampel.struct.UnitResult import UnitResult
from ampel.types import T3Send, UBson
from ampel.view.SnapView import SnapView


class T3BufferExporterUnit(AbsT3Unit[SnapView]):

	# None means stdout
	fd: None | str = None
	binary: bool = False
	verbose: bool = True
	id_mapper: None | AbsIdMapper = None
	txt_options: TextExportOptions = TextExportOptions()


	def process(self,
		gen: Generator[SnapView, T3Send, None],
		t3s: T3Store
	) -> UBson | UnitResult:

		fd, close_fd = get_fd(self.binary, self.fd, True)

		if self.binary:
			bin_export(
				fd, # type: ignore[arg-type]
				self.serialize_views(gen), self.id_mapper, close_fd,
				logger = self.logger if self.verbose else None
			)

		else:
			txt_export(
				fd, # type: ignore[arg-type]
				self.serialize_views(gen), self.id_mapper,
				**self.txt_options.dict(), close_fd = close_fd,
				logger = self.logger if self.verbose else None
			)
		
		return None


	def serialize_views(self,
		gen: Generator[SnapView, T3Send, None],
	) -> Generator[AmpelBuffer, None, None]:
		for sv in gen:
			yield serialize(sv)

serialize = TypeAdapter(SnapView).dump_python



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/var/DBLoggingHandler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.12.2017
# Last Modified Date:  15.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import socket
import struct
from logging import LogRecord
from traceback import format_exc
from typing import TYPE_CHECKING

from bson import ObjectId
from pymongo.errors import BulkWriteError

from ampel.base.AmpelUnit import AmpelUnit
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.AmpelLoggingError import AmpelLoggingError
from ampel.log.LightLogRecord import LightLogRecord
from ampel.log.LogFlag import LogFlag
from ampel.log.LoggingErrorReporter import LoggingErrorReporter
from ampel.log.utils import log_exception, report_exception
from ampel.util.collections import try_reduce
from ampel.util.mappings import compare_dict_values
from ampel.util.pretty import prettyjson

if TYPE_CHECKING:
	from ampel.core.AmpelDB import AmpelDB


# http://isthe.com/chongo/tech/comp/fnv/index.html#FNV-1a
def _fnv_1a_24(data, _ord=lambda x: x):
	"""FNV-1a 24 bit hash"""
	# http://www.isthe.com/chongo/tech/comp/fnv/index.html#xor-fold
	# Start with FNV-1a 32 bit.
	hash_size = 2 ** 32
	fnv_32_prime = 16777619
	fnv_1a_hash = 2166136261  # 32-bit FNV-1 offset basis
	for elt in data:
		fnv_1a_hash = fnv_1a_hash ^ _ord(elt)
		fnv_1a_hash = (fnv_1a_hash * fnv_32_prime) % hash_size

	# xor-fold the result to 24 bit.
	return (fnv_1a_hash >> 24) ^ (fnv_1a_hash & 0xffffff)


def _machine_bytes():
	"""Get the machine portion of an ObjectId.
	"""
	# gethostname() returns a unicode string in python 3.x
	# We only need 3 bytes, and _fnv_1a_24 returns a 24 bit integer.
	# Remove the padding byte.
	return struct.pack("<I", _fnv_1a_24(socket.gethostname().encode()))[:3]


class DBLoggingHandler(AmpelUnit):
	""" Saves log events into mongo database """

	__slots__ = "prev_record", "fields_check", "log_dicts", "oid_middle", "warn_lvl"

	level: int
	col_name: str = "log"
	aggregate_interval: float = 1.
	expand_extra: bool = False # Whether extra keys should be moved as root LogDocument keys
	flush_len: int = 1000
	auto_flush: bool = False
	log_provenance: bool = True


	def __init__(self, ampel_db: 'AmpelDB', run_id: int, **kwargs) -> None:
		"""
		:param col_name: name of db collection to use (default: 'logs' in database Ampel_var)
		:param aggregate_interval: logs with similar attributes (log level, possibly tranId & channels) \
		are aggregated in one document instead of being split into several documents (spares some index RAM). \
		*aggregate_interval* is the max interval of time in seconds during which log aggregation takes place. \
		Beyond this value, a new log document is created no matter what. This parameter thus impacts logging time granularity.
		:param flush_len: How many log documents should be kept in memory before attempting a database bulk_write operation.
		"""

		super().__init__(**kwargs)
		self._ampel_db = ampel_db
		self.run_id = run_id

		self.log_dicts: list[dict] = []
		self.prev_record: None | LightLogRecord | LogRecord = None
		self.fields_check = ['extra', 'stock', 'channel']
		self.warn_lvl = LogFlag.WARNING

		# Get reference to pymongo collection
		self.col = ampel_db.get_collection(self.col_name)

		# ObjectID middle: 3 bytes machine + 2 bytes encoding the last 4 digits of run_id (unique)
		# NB: pid is not always unique if running in a jail or container
		self.oid_middle = _machine_bytes() + int(str(self.run_id)[-4:]).to_bytes(2, 'big')


	def handle(self, record: LightLogRecord | LogRecord) -> None:
		""" :raises AmpelLoggingError: on error """

		rd = record.__dict__
		prev_record = self.prev_record

		try:

			# Same flag, date (+- 1 sec), tran_id and chans
			if (
				prev_record and
				(record.name is None or record.name == prev_record.name) and
				record.levelno <= prev_record.levelno and
				record.created - prev_record.created < self.aggregate_interval and
				compare_dict_values(rd, prev_record.__dict__, self.fields_check)
			):

				prev_dict = self.log_dicts[-1]

				if 'm' not in prev_dict:
					prev_dict['m'] = "None log entry with identical fields repeated twice"
					return

				if isinstance(prev_dict['m'], list):
					prev_dict['m'].append(record.msg)
				else:
					prev_dict['m'] = [prev_dict['m'], record.msg]

			else:

				if self.auto_flush and len(self.log_dicts) > self.flush_len:
					self.flush()

				# Treat SHOUT msg as INFO msg (and try again to concatenate)
				if record.levelno == LogFlag.SHOUT:
					if isinstance(record, LightLogRecord):
						new_rec = LightLogRecord(name=0, msg=None, levelno=0)
					else:
						new_rec = LogRecord(name=None, pathname=None, level=None, # type: ignore[arg-type, assignment]
							lineno=None, exc_info=None, msg=None, args=None) # type: ignore[arg-type]
					for k, v in record.__dict__.items():
						new_rec.__dict__[k] = v
					new_rec.levelno = LogFlag.INFO
					self.handle(new_rec)
					return

				# Generate object id with log record.created as current time
				with ObjectId._inc_lock:  # noqa: SLF001
					oid = struct.pack(">i", int(record.created)) + \
						self.oid_middle + struct.pack(">i", ObjectId._inc)[1:4]  # noqa: SLF001
					ObjectId._inc = (ObjectId._inc + 1) % 0xFFFFFF # limit result to 32bits  # noqa: SLF001

				if 'extra' in rd:
					for k in ('_id', 'r', 'f'):
						if k in rd['extra']:
							del rd['extra'][k]
					if self.expand_extra:
						ldict = {
							'_id': ObjectId(oid=oid),
							'r': self.run_id,
							'f': record.levelno,
							**rd['extra']
						}
					else:
						ldict = {
							'_id': ObjectId(oid=oid),
							'r': self.run_id,
							'f': record.levelno,
							'x': rd['extra']
						}
				else:
					ldict = {
						'_id': ObjectId(oid=oid),
						'r': self.run_id,
						'f': record.levelno
					}

				if 'stock' in rd:
					ldict['s'] = rd['stock']

				if 'channel' in rd:
					ldict['c'] = try_reduce(rd['channel'])

				if record.filename and (record.levelno > self.warn_lvl or self.log_provenance):
					ldict['p'] = [record.filename.replace('.py', ''), record.lineno]

				if hasattr(record, 'unit'):
					ldict['u'] = record.unit

				if record.msg:
					ldict['m'] = record.msg

				self.log_dicts.append(ldict)
				self.prev_record = record

		except Exception as e:

			logger = AmpelLogger.get_logger()

			# Print log stack using std logging
			log_exception(logger, e, msg="Exception:")

			try:
				# This will fail as well if we have DB connectivity issues
				report_exception(self._ampel_db, logger, e)
			except Exception as ee:
				log_exception(
					logger, ee, last=True,
					msg="Could not update troubles collection as well (DB offline?)"
				)

			logger.error("DB log flushing error, un-flushed (json) logs below.")
			logger.error("*" * 52)

			for d in self.log_dicts:
				logger.error(prettyjson(d))
			logger.error("#" * 52)

			raise AmpelLoggingError from None


	def get_run_id(self) -> int:
		return self.run_id


	def check_flush(self) -> None:
		if len(self.log_dicts) > self.flush_len:
			self.flush()


	def break_aggregation(self) -> None:
		self.prev_record = None


	def flush(self) -> None:
		""" :raises AmpelLoggingError: on error """

		# No log entries
		if not self.log_dicts:
			return

		# save referenced log entries locally
		dicts = self.log_dicts

		# clear referenced log entries
		self.log_dicts = []
		self.prev_record = None

		try:
			# pymongo drops the GIL while sending and receiving data over the network
			self.col.insert_many(dicts, ordered=False)

		except BulkWriteError as bwe:
			if self.handle_bulk_write_error(bwe):
				raise AmpelLoggingError from None

		except Exception as e:
			LoggingErrorReporter.report(self, e)
			# If we can no longer keep track of what Ampel is doing,
			# better raise Exception to stop processing
			raise AmpelLoggingError from None


	def handle_bulk_write_error(self, bwe: BulkWriteError) -> bool:
		"""
		:returns: true if error could not be handled properly
		"""
		# We try here to handle duplicate key errors which can occur in rare scenarios.
		# We loop through each element in the returned writeErrors and
		# see for the E11000 cases if - in the DB - the runId of the log record
		# with conflicting ObjectId is the same as the one from the current instance.
		# If the runId are the same, then the conflicting log entry originates from
		# the current instance and we trust it as being the same than the one in the buffer
		# Further equality checks (msg content) could be done but we don't do them for now.

		try:

			from ampel.log.utils import (  # noqa: PLC0415
				convert_dollars,
				get_tier_from_log_flags,
				log_exception,
			)
			raise_exc = False

			for err_dict in bwe.details.get('writeErrors', []):

				# This log entry was already inserted into DB
				# 'code': 11000, 'errmsg': 'E11000 duplicate key error collection: ...
				if err_dict.get("code") == 11000:

					# Fetch corresponding log entry from DB
					db_rec = next(self.col.find({'_id': err_dict['op']['_id']}))

					# if runIds are equal, just print out a feedback
					if db_rec['run'] == self.run_id:
						print("Disregardable E11000: {}".format(err_dict['op']))  # noqa: T201

					# Otherwise print error and create trouble doc
					else:

						raise_exc = True

						print("CRITICAL: OID collision occured between two different log entries")  # noqa: T201
						print("Current process:")  # noqa: T201
						print(err_dict['op'])  # noqa: T201
						print("In DB:")  # noqa: T201
						print(db_rec)  # noqa: T201

						self._ampel_db.get_collection('trouble').insert_one(
							{
								'tier': get_tier_from_log_flags(db_rec['f']),
								'location': 'DBLoggingHandler',
								'msg': "OID collision occured between two different log entries",
								'ownLogEntry': str(err_dict['op']),
								'existingLogEntry': str(db_rec)
							}
						)
				else:

					raise_exc = True
					print(f"writeError dict entry: {err_dict}")  # noqa: T201

					self._ampel_db.get_collection('trouble').insert_one(
						{
							'location': 'DBLoggingHandler',
							'msg': "non-E11000 writeError occured",
							'errorDict': convert_dollars(err_dict)
						}
					)

			return raise_exc

		# bad day
		except Exception:

			try:
				self._ampel_db.get_collection('trouble').insert_one(
					{
						'location': 'DBLoggingHandler',
						'msg': "Exception occured in handle_bulk_write_error",
						'exception': format_exc().replace("\"", "'").split("\n")
					}
				)
			except Exception as another_exc:
				from ampel.log.AmpelLogger import AmpelLogger  # noqa: PLC0415
				log_exception(
					AmpelLogger.get_logger(),
					another_exc
				)

			from ampel.log.LoggingErrorReporter import (  # noqa: PLC0415
				LoggingErrorReporter,
			)
			LoggingErrorReporter.report(self, bwe, bwe.details)

			return True



import pytest
from prometheus_client.metrics import MetricWrapperBase

from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry, reset_registry


@pytest.fixture
def empty_registry():
    prev = AmpelMetricsRegistry._registry
    AmpelMetricsRegistry._registry = None
    yield AmpelMetricsRegistry.registry()
    AmpelMetricsRegistry._registry = prev


def test_autoname(empty_registry):
    assert (
        AmpelMetricsRegistry.counter(
            "foo", "foos", unit="blarghs", subsystem="blah"
        )._name
        == "ampel_blah_foo_blarghs"
    )


def verify(metric: MetricWrapperBase, value: int):
    for sample in next(iter(metric.collect())).samples:
        if not sample.name.endswith("_created"):
            assert sample.value == value


def test_reset_counter(empty_registry):
    metric = AmpelMetricsRegistry.counter("foo", "foos")
    verify(metric, 0)
    metric.inc(1)
    verify(metric, 1)
    reset_registry(empty_registry)
    verify(metric, 0)


def test_reset_summary(empty_registry):
    metric = AmpelMetricsRegistry.summary("foo", "foos")
    verify(metric, 0)
    metric.observe(1)
    verify(metric, 1)
    reset_registry(empty_registry)
    verify(metric, 0)


def test_reset_histogram(empty_registry):
    metric = AmpelMetricsRegistry.histogram("foo", "foos", buckets=[1, float("inf")])
    verify(metric, 0)
    metric.observe(1)
    verify(metric, 1)
    reset_registry(empty_registry)
    verify(metric, 0)


def test_reset_gauge(empty_registry):
    metric = AmpelMetricsRegistry.gauge("foo", "foos")
    verify(metric, 0)
    metric.set(1)
    verify(metric, 1)
    reset_registry(empty_registry)
    verify(metric, 0)



from collections.abc import Sequence
from typing import Protocol

from bson import ObjectId

from ampel.content.JournalRecord import JournalRecord
from ampel.content.StockDocument import StockDocument
from ampel.enum.JournalActionCode import JournalActionCode
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.types import ChannelId, StockId, Tag


class StockUpdaterProtocol(Protocol):
	def add_journal_record(self,
		stock: StockId | Sequence[StockId],
		jattrs: None | JournalAttributes = None,
		tag: None | Tag | Sequence[Tag] = None,
		name: None | str | Sequence[str] = None,
		trace_id: None | dict[str, int] = None,
		action_code: None | JournalActionCode = None,
		doc_id: None | ObjectId = None,
		unit: None | int | str = None,
		channel: None | ChannelId | Sequence[ChannelId] = None,
		now: None | int | float = None
	) -> JournalRecord:
		"""
		Add a journal record to the stock document(s) identified by the input stock id(s)
		"""

	def add_name(self, stock: StockId, name: str | Sequence[str]) -> None: ...

	def add_tag(self,
		stock: StockId | Sequence[StockId],
		tag: Tag | Sequence[Tag]
	) -> None: ...

class StockIngesterProtocol(Protocol):
    def ingest(self, doc: StockDocument) -> None:
        ...
    
    @property
    def update(self) -> StockUpdaterProtocol:
        ...


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/load/T3SimpleDataLoader.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.12.2019
# Last Modified Date:  29.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Iterable, Iterator
from typing import ClassVar

from bson.codec_options import CodecOptions

from ampel.abstract.AbsT3Loader import AbsT3Loader
from ampel.mongo.view.FrozenValuesDict import FrozenValuesDict
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.types import StockId, StrictIterable


class T3SimpleDataLoader(AbsT3Loader):
	"""Load all requested documents for the selected stocks"""

	codec_options: ClassVar[None | CodecOptions] = CodecOptions(document_class=FrozenValuesDict)

	def load(self,
		stock_ids: StockId | Iterator[StockId] | StrictIterable[StockId]
	) -> Iterable[AmpelBuffer]:

		return self.data_loader.load(
			stock_ids = stock_ids,
			directives = self.directives,
			channel = self.channel,
			codec_options = self.codec_options,
			logger = self.logger
		)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/JobTaskModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                05.04.2023
# Last Modified Date:  05.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Literal

from ampel.base.AmpelBaseModel import AmpelBaseModel


class JobTaskModel(AmpelBaseModel):
	"""
	A note about templates definitions:
	1) str | list[str]: templates are applied prior to job run
	2) same applies for dict[Literal['pre'], str | list[str]] -> 1) is just a shorthand for 2)
	3) dict[Literal['live'], str | list[str]] processes templates during job run,
	  just before the task is executed. It is required when templates
	  necessitate run-time information generated by previous tasks.
	The same template can be used as a 'pre' or 'live' template depending on the situation.
	For example, hash_t2_config can be used prior to job run for simple jobs.
	However, if run-time aliases are used, hash_t2_config should be requested as a 'live' template
	(just after resolve_run_time_aliases) as the hashing of unit configurations
	must occur after the inclusion of run-time information.
	"""

	# allow arbitrary content for templates
	model_config = {
		"extra": "allow"
	}

	title: None | str
	template: None | str | list[str] | dict[Literal['pre', 'live'], None | str | list[str]]
	unit: None | str # allow none as unit shall be set by template
	config: None | int | str | dict[str, Any]
	multiplier: int = 1

	def get_multiplier(self) -> int:
		return self.multiplier



from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from functools import partial
from typing import Any, Literal

from pymongo.client_session import ClientSession
from pymongo.errors import OperationFailure
from pymongo.read_concern import ReadConcern
from pymongo.read_preferences import ReadPreference
from pymongo.write_concern import WriteConcern

from ampel.abstract.AbsOpsUnit import AbsOpsUnit
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.content.StockDocument import StockDocument
from ampel.core.AmpelDB import AmpelDB
from ampel.log.utils import safe_query_dict
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.time.TimeConstraintModel import TimeConstraintModel
from ampel.mongo.query.stock import build_stock_query
from ampel.types import ChannelId, StockId, Tag
from ampel.util.collections import get_chunks


class WriteConcernModel(AmpelBaseModel):
    w: Literal["majority"] | int = 1
    j: bool = True
    wtimeout: int = 0


class StockSelectionModel(AmpelBaseModel):

    #: Select by creation time
    created: None | TimeConstraintModel = None

    #: Select by modification time
    updated: None | TimeConstraintModel = None

    #: Select by channel
    channel: None | ChannelId | AnyOf[ChannelId] | AllOf[ChannelId] | OneOf[
        ChannelId
    ] = None

    #: Select by tag
    tag: None | dict[
        Literal["with", "without"], Tag | AllOf[Tag] | AnyOf[Tag] | OneOf[Tag] | dict
    ] = None

    #: Custom selection (ex: {'run': {'$gt': 10}})
    custom: None | dict[str, Any] = None

    def get_query(self, db: AmpelDB, now: None | datetime) -> dict[str, Any]:

        # Build query for matching transients using criteria defined in config
        match_query = build_stock_query(
            channel=self.channel,
            tag=self.tag,
            time_created=self.created.get_query_model(db=db, now=now)
            if self.created
            else None,
            time_updated=self.updated.get_query_model(db=db, now=now)
            if self.updated
            else None,
        )

        if self.custom:
            match_query.update(self.custom)

        return match_query


class MongoStockDeleter(AbsOpsUnit):
    """
    Delete all documents associated with a set of stocks
    """

    #: number of stocks to purge in a single transaction
    chunk_size: int = 1000
    #: when to consider "now"
    now: Literal["latest_stock", "now"] = "latest_stock"
    #: stocks to delete
    delete: StockSelectionModel
    #: roll back each transaction before it can be committed
    dry_run: bool = True
    #: write concern to use for transaction
    write_concern: WriteConcernModel = WriteConcernModel(w=1, j=True, wtimeout=0)
    causal_consistency: bool = True
    retry: int = 0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._collections = {"stock": self.context.db.get_collection("stock"),} | {
            f"t{tier}": self.context.db.get_collection(f"t{tier}") for tier in range(3)
        }

    def _purge_chunk_in_transaction(
        self, session: ClientSession, stock_ids: list[StockId]
    ) -> dict[str, int]:

        deleted = {k: 0 for k in self._collections}

        doc_match = {"stock": {"$in": stock_ids}}

        def delete(k):
            result = self._collections[k].delete_many(doc_match, session=session)
            self.logger.debug(
                None,
                extra={
                    "col": k,
                    "deleted": result.deleted_count if result.acknowledged else 0,
                },
            )
            return k, result.deleted_count if result.acknowledged else 0

        with ThreadPoolExecutor(max_workers=4) as executor:
            for k, v in executor.map(
                delete, [f"t{tier}" for tier in range(2, -1, -1)] + ["stock"]
            ):
                deleted[k] += v

        if self.dry_run:
            self.logger.debug("Rollback")
            session.abort_transaction()
        else:
            self.logger.debug("Commit")

        return deleted

    def _purge_chunk(
        self, session: ClientSession, stock_ids: list[StockId]
    ) -> dict[str, int]:
        for attempt in range(self.retry + 1):
            try:
                return session.with_transaction(
                    partial(self._purge_chunk_in_transaction, stock_ids=stock_ids),
                    write_concern=WriteConcern(**self.write_concern.dict()),
                )
            except OperationFailure as exc:  # noqa: PERF203
                # operation was interrupted because the transaction exceeded the configured 'transactionLifetimeLimitSeconds'
                if attempt == self.retry or exc.code != 290:
                    raise
                continue
        # unreachable
        raise NotImplementedError()

    def run(self, beacon: None | dict[str, Any] = None) -> None | dict[str, Any]:

        if self.now == "now":
            now = datetime.now(tz=timezone.utc)
        elif self.now == "latest_stock":
            # use the timestamp of the most recently inserted stock as a marker for "now"
            try:
                latest_stock: StockDocument = next(
                    self._collections["stock"]
                    .find({}, {"ts": 1})
                    .sort([("_id", -1)])
                    .limit(1)
                )
            except StopIteration:
                return None
            now = datetime.fromtimestamp(latest_stock["ts"]["any"]["tied"], tz=timezone.utc)
            self.logger.info(f"Last stock inserted {now} ({now.timestamp():.0f})")

        # TODO: specify conditions for stocks to delete (e.g. per channel), and then $and them
        stock_match = self.delete.get_query(self.context.db, now)

        if self.logger.verbose:
            self.logger.info(
                f"Purging {self._collections['stock'].count_documents(stock_match)} of {self._collections['stock'].estimated_document_count()} stocks",
                extra=safe_query_dict(stock_match),
            )
        else:
            self.logger.info("Purging stocks", extra=safe_query_dict(stock_match))

        deleted = {k: 0 for k in self._collections}

        with self._collections["stock"].database.client.start_session(
            causal_consistency=self.causal_consistency
        ) as session:
            deleted_stocks = 0
            for docs in get_chunks(
                self._collections["stock"]
                .with_options(
                    read_concern=ReadConcern(level="local"),
                    read_preference=ReadPreference.SECONDARY_PREFERRED,
                )
                .find(
                    stock_match,
                    {"stock": 1},
                    session=session,
                )
                .sort("stock", 1),
                self.chunk_size,
            ):
                self.logger.debug("Purging chunk")
                deleted_in_chunk = self._purge_chunk(
                    session, stock_ids=[doc["stock"] for doc in docs]
                )
                for k, v in deleted_in_chunk.items():
                    deleted[k] += v
                deleted_stocks += len(docs)
                self.logger.info(f"Purged {deleted_stocks} stocks", extra=deleted)

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/filter/T3AmpelBufferFilter.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.01.2020
# Last Modified Date:  18.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import collections
from collections.abc import Iterable, Sequence
from typing import Literal, get_args

from ampel.abstract.AbsT3Filter import AbsT3Filter
from ampel.aux.filter.AbsLogicOperatorFilter import AbsLogicOperatorFilter
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.log.AmpelLogger import AmpelLogger
from ampel.model.operator.AllOf import AllOf
from ampel.model.operator.AnyOf import AnyOf
from ampel.model.operator.OneOf import OneOf
from ampel.model.UnitModel import UnitModel
from ampel.struct.AmpelBuffer import AmpelBuffer, BufferKey
from ampel.types import ChannelId

channel_id = get_args(ChannelId)


class FilterBlock(AmpelBaseModel):

	data: BufferKey | Literal['journal']
	filter: AbsLogicOperatorFilter
	include: bool = True

	def __init__(self, **kwargs):
		AmpelBaseModel.__init__(self, **kwargs)
		typ = "include" if self.include else "exclude"
		self.descr = f"{self.filter.__class__.__name__}[target={self.data}, on_match={typ}]"


class T3AmpelBufferFilter(AbsT3Filter):

	class FilterModel(AmpelBaseModel):
		data: BufferKey | Literal['journal']
		filter: UnitModel
		on_match: Literal['include', 'exclude'] = 'include'

	logger: AmpelLogger
	filters: Sequence[FilterModel] = []
	channel: None | ChannelId | AllOf[ChannelId] | AnyOf[ChannelId] | OneOf[ChannelId] = None

	def __init__(self, **kwargs) -> None:

		# Kulanz
		if 'filters' in kwargs and not isinstance(kwargs['filters'], collections.abc.Sequence):
			kwargs['filters'] = [kwargs['filters']]

		super().__init__(**kwargs)
		self.filter_blocks: list[FilterBlock] = []

		for f in self.filters:

			self.filter_blocks.append(
				FilterBlock(
					data = f.data,
					filter = AuxUnitRegister.new_unit(f.filter),
					include = f.on_match == "include"
				)
			)


	def filter(self, it: Iterable[AmpelBuffer]) -> Sequence[AmpelBuffer]:

		debug = self.logger.verbose > 1
		ret: list[AmpelBuffer]

		if self.channel:

			if isinstance(self.channel, channel_id):
				ret = [
					ab for ab in it
					if ab['stock']['channel'] == self.channel or # type: ignore[index]
					self.channel in ab['stock']['channel'] # type: ignore[index]
				]


			elif isinstance(self.channel, AnyOf):
				ret = []
				for ab in it:
					if 'stock' not in ab or not ab['stock']:
						continue
					for chan in self.channel.any_of:
						if ab['stock']['channel'] == chan or chan in ab['stock']['channel']:
							ret.append(ab)
							break

			elif isinstance(self.channel, AllOf):
				ret = []
				for ab in it:
					if 'stock' not in ab or not ab['stock']:
						continue
					if all([chan in ab['stock']['channel'] for chan in self.channel.all_of]):
						ret.append(ab)

			elif isinstance(self.channel, OneOf):
				ret = []
				for ab in it:
					if 'stock' not in ab or not ab['stock']:
						continue
					if len(ab['stock']['channel']) == 1:
						for chan in self.channel.one_of:
							if ab['stock']['channel'][0] == chan:
								ret.append(ab)
								break
			else:
				raise ValueError("Unrecognized parameter")

			# if debug and len(ret) != len(it):
			#	self.logger.debug(f"AmpelBuffer stock channel filter: in: {len(it)}, out: {len(ret)}")

		else:
			ret = list(it)

		for fb in self.filter_blocks:

			jfilter = fb.data == 'journal'

			# Using the old trick of iterating backwards to remove list element while iterating over list
			for i in range(len(ret) - 1, -1, -1):

				abuf = ret[i]

				if jfilter:
					if 'stock' not in abuf:
						raise ValueError("A StockDocument is required to filter journal entries")
					in_arr = abuf['stock']['journal'] # type: ignore[index]
				else:
					in_arr = abuf[fb.data] # type: ignore[typeddict-item]

				fres = fb.filter.apply(in_arr)

				if debug:
					self.logger.debug(f"{fb.descr} result: in: {len(in_arr)}, out: {len(fres)}")

				if fb.include:
					if not fres: # matched elements should be include but no match
						ret.pop(i)
				elif fres: # elements were matched but on_match is exclude
					ret.pop(i)

		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/stage/project/T3BaseProjector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.01.2020
# Last Modified Date:  15.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Callable, Iterable, Sequence
from typing import Any

from ampel.abstract.AbsApplicable import AbsApplicable
from ampel.abstract.AbsT3Projector import AbsT3Projector
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.log import VERBOSE, AmpelLogger
from ampel.model.UnitModel import UnitModel
from ampel.struct.AmpelBuffer import AmpelBuffer, BufferKey
from ampel.types import Traceless


class T3BaseProjector(AbsT3Projector):
	"""
	Creates SnapView(s) based on AmpelBuffer(s)
	
	AmpelBuffer attributes are selected according to 'filters' content:
	
	- missing fields will not be included
	  (for example, if 'names' is absent in filters dict,
	  the field 'names' of all snapviews will be None)
	- a "pass-through" behavior occurs when a dict value are None.
	  (for example, if 'names' is set to None, snapviews instances
	  will contain the all the names of the corresponding AmpelBuffer)
	"""

	class ClassModel(AmpelBaseModel):
		key: BufferKey
		model: UnitModel

	class FuncModel(AmpelBaseModel):
		key: BufferKey
		func: Callable[[Any], Any]

	class FilterOutModel(AmpelBaseModel):
		discard: BufferKey


	logger: Traceless[AmpelLogger]

	# Projections can yield empty buffers, especially if the previous 'filter' stage was not used.
	# Ex: # {
	#	'_id': 318144, 'channel': None, 'created': {}, 'journal': (),
	# 	'updated': {}, 'name': ('ZTF17aaabdku',), 'tag': ('ZTF',)
	# }
	# This parameter determines if those empty buffers should be removed.
	# Detection is based on the None value of <buffer>['stock']['channel']
	# and thus requires 'stock' documents to loaded which should be the most frequent case
	remove_empty: bool = True

	# Modify/delete dict keys/values
	field_projectors: list[ClassModel | FuncModel] = []


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)

		# List matchers
		self.projectors: dict[BufferKey, None | list[Callable[[Any], Any]]] = {}

		for fp in self.field_projectors:
			if isinstance(fp, self.ClassModel):
				self.add_class_projector(fp)
			elif isinstance(fp, self.FuncModel):
				self.add_func_projector(fp.key, fp.func)
			elif isinstance(fp, self.FilterOutModel):
				self.projectors[fp.discard] = None

		self.pass_through_keys: set[BufferKey] = {"stock", "t0", "t1", "t2", "logs", "extra"} - self.projectors.keys() # type: ignore[assignment]


	def add_class_projector(self, cm: ClassModel, first: bool = False) -> None:

		# Instantiate field projector units (auxiliary units)
		unit = AuxUnitRegister.new_unit(
			model = cm.model,
			sub_type = AbsApplicable,
			logger = self.logger
		)

		self.add_func_projector(cm.key, unit.apply, first=first)


	def add_func_projector(self,
		key: BufferKey, func: Callable, first: bool = False
	) -> None:
		"""
		A matcher rejects/accepts dict instances from a list referenced by the given key.
		For example, a ampel_buffer intance can contain many t2 documents (ampel_buffer dict key "t2").
		The matchers will determine which dict from the list should be kept
		(a typical case would be to exclude dicts based on channel associations)

		:param first: whether field matchers should be put first in the list of matchers for the provided key
		:raise: ValueError if requested field projector unit / configs are invalid
		"""

		if self.logger.verbose > 1:
			self.logger.debug(f"Adding function '{func.__name__}' as field projector for '{key}'")

		# Upate internal set
		self.pass_through_keys -= set([key])

		# New projector for this key
		if key in self.projectors:

			if self.projectors[key] is None:
				raise ValueError(
					f"Cannot setup projector for key '{key}' as the "
					f"null projector was requested for this key"
				)
			# Put new projector first in the list for this key
			if first:
				self.projectors[key].insert(0, func) # type: ignore[union-attr]
			else:
				self.projectors[key].append(func) # type: ignore[union-attr]

		# No field projector was setup for this key yet
		else:
			self.projectors[key] = [func]


	def project(self, ampel_buffer: Iterable[AmpelBuffer]) -> Sequence[AmpelBuffer]:

		# micro optimization
		projectors = self.projectors
		pass_through_keys = self.pass_through_keys

		ret: list[AmpelBuffer] = []

		for abuf in ampel_buffer:

			new_buf = AmpelBuffer(id=abuf["id"])

			for k in pass_through_keys:
				if k in abuf:
					new_buf[k] = abuf[k]

			# Loop through field projectors defined in config
			for k in projectors:

				# Last condition handles the 'null projector'
				if k in abuf and abuf[k] and projectors[k]:

					new_buf[k] = abuf[k]
					for p in projectors[k]: # type: ignore[union-attr]
						new_buf[k] = p(new_buf[k])
						if not new_buf[k]:
							break

					if self.logger.verbose:

						if isinstance(abuf[k], list | tuple):
							self.logger.log(VERBOSE, None,
								extra={'projection': k, 'in': len(abuf[k]), 'out': len(new_buf[k])} # type: ignore[arg-type]
							)
						elif k == 'stock':
							self.logger.log(VERBOSE, None,
								extra={
									'projection': k,
									'journal_in': len(abuf[k]['journal']),     # type: ignore[index]
									'journal_out': len(new_buf[k]['journal'])  # type: ignore[index]
								}
							)

						#for d in [abuf[k]] if k=='stock' else abuf:
						if self.logger.verbose > 1:
							self.logger.debug(None, extra={'projection': k, 'input': abuf[k]})
							self.logger.debug(None, extra={'projection': k, 'output': new_buf[k]})

			# If the previous stage did not filter properly (or at all), projections can yield empty buffers like this:
			# By default, we remove those
			# Note that it actually requires 'stock' documents to be requested/loaded which should be almost always the case
			if self.remove_empty and 'stock' in new_buf and not new_buf['stock']['channel']: # type: ignore[index]
				if self.logger.verbose > 1:
					self.logger.debug(f"Removing {new_buf['id']!r} emptied by projection")
				continue

			ret.append(new_buf)

		return ret



from collections.abc import Callable
from dataclasses import dataclass
from typing import TypeVar

from ampel.abstract.AbsContextManager import AbsContextManager
from ampel.base.AmpelUnit import AmpelUnit
from ampel.base.decorator import abstractmethod
from ampel.content.DataPoint import DataPoint
from ampel.content.StockDocument import StockDocument
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document

_T = TypeVar("_T", StockDocument, DataPoint, T1Document, T2Document)

class AbsProducer(AbsContextManager, AmpelUnit, abstract=True):

    @dataclass
    class Item:
        """
        A bundle of documents that all belong to the same context (e.g. an alert)
        """

        stock: list[StockDocument]
        t0: list[DataPoint]
        t1: list[T1Document]
        t2: list[T2Document]

        @classmethod
        def new(cls) -> "AbsProducer.Item":
            return cls([], [], [], [])
        
        def __bool__(self) -> bool:
            return any((self.stock, self.t0, self.t1, self.t2))

    @abstractmethod
    def produce(
        self, item: Item, delivery_callback: None | Callable[[], None]
    ) -> None: ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/LightLogRecord.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                22.04.2020
# Last Modified Date:  15.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from typing import Any

from ampel.types import ChannelId, StockId


class LightLogRecord:
	"""
	LogRecord class similar to the one provided by standard 'logging' module but faster.

	In []: %timeit LogRecord(name=12, pathname=None, level=12, lineno=12, exc_info=None, msg=None, args=None)
	Out[]: 4.7 µs ± 77.9 ns per loop (mean ± std. dev. of 7 runs, 100000 loops each)

	In []: %timeit LightLogRecord(name=12, levelno=1, msg=None)
	Out[]: 657 ns ± 6.51 ns per loop (mean ± std. dev. of 7 runs, 1000000 loops each)
	"""

	name: int | str
	levelno: int
	msg: None | str | dict[str, Any]
	channel: None | ChannelId | list[ChannelId]
	stock: None | StockId
	extra: None | dict[str, Any]
	unit: None | str

	def __init__(self,
		name: int | str,
		levelno: int,
		msg: None | str | dict[str, Any] = None
	) -> None:

		d = self.__dict__
		d['name'] = name
		d['created'] = time()
		d['levelno'] = levelno
		d['msg'] = msg

	def getMessage(self):
		return self.msg

	def __getattr__(self, k):
		return self.__dict__.get(k, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/complement/T3LogsAppender.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.03.2021
# Last Modified Date:  28.02.2022
# Last Modified By:    Marcus Fenner <mf@physik.hu-berlin.de>

from collections.abc import Iterable
from datetime import datetime, timezone
from typing import Any

from bson.objectid import ObjectId

from ampel.abstract.AbsBufferComplement import AbsBufferComplement
from ampel.log.utils import safe_query_dict
from ampel.mongo.query.var.LogsLoader import LogsLoader
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.struct.T3Store import T3Store


class T3LogsAppender(AbsBufferComplement):

	use_last_run: bool = True
	logs_loader_conf: dict[str, Any] = {}

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.log_loader = LogsLoader(**self.logs_loader_conf, read_only=True)
		self.col = self.context.db.get_collection('log')


	def complement(self, it: Iterable[AmpelBuffer], t3s: T3Store) -> None:

		query: dict[str, Any] = {'s': {'$in': [el['id'] for el in it]}}

		if t3s.session and self.use_last_run and t3s.session.get('last_run'):
			query['_id'] = {
				'$gte': ObjectId.from_datetime(
					datetime.fromtimestamp(
						t3s.session['last_run'],
						tz=timezone.utc,
					)
				)
			}

		logs = list(
			self.log_loader.fetch_logs(self.col, query)
		)

		self.logger.debug(
			f"Log query returned {len(logs)} result(s)",
			extra=safe_query_dict(query)
		)

		if not logs:
			return

		for ab in it:
			if 'logs' not in ab or ab['logs'] is None:
				ab['logs'] = []
			for l in logs:
				if ab['id'] == l['s']:
					ab['logs'].append(l) # type: ignore[union-attr]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/T3DefaultBufferSupplier.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.07.2021
# Last Modified Date:  13.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator, Sequence

from ampel.abstract.AbsBufferComplement import AbsBufferComplement
from ampel.abstract.AbsT3Loader import AbsT3Loader
from ampel.abstract.AbsT3Selector import AbsT3Selector
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.model.UnitModel import UnitModel
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.struct.T3Store import T3Store
from ampel.util.collections import get_chunks as chunks_func


class T3DefaultBufferSupplier(AbsT3Supplier[Generator[AmpelBuffer, None, None]]):
	"""
	Default T3 supplier that:
	- matches stock ids based on the stock collection (section 'select'),
	- loads AmpelBuffer content as defined in the 'load' section
	- potentially updates AmpelBuffer through the 'complement' section
	"""

	#: Select stocks
	select: UnitModel

	#: Fill :class:`~ampel.core.AmpelBuffer.AmpelBuffer` for each selected stock
	load: UnitModel

	#: Add external information to each :class:`~ampel.core.AmpelBuffer.AmpelBuffer`.
	complement: None | Sequence[UnitModel]

	#: number of stocks to load at once. Set to 0 to disable chunking
	chunk_size: int = 1000


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)

		# stock selection
		#################

		# Spawn and run a new selector instance
		# fetch() returns an iterable (often a pymongo cursor)
		self.selector = self.context.loader.new_context_unit(
			model = self.select,
			context = self.context,
			sub_type = AbsT3Selector,
			logger = self.logger
		)

		# Content loader
		################

		# Spawn requested content loader
		self.data_loader = self.context.loader \
			.new_context_unit(
				model = self.load,
				context = self.context,
				sub_type = AbsT3Loader,
				logger = self.logger
			)

		# Content complementer
		######################

		if self.complement:

			# Spawn requested snapdata complementers
			self.complementers: None | list[AbsBufferComplement] = [
				self.context.loader \
					.new_context_unit(
						model = conf_el,
						context = self.context,
						sub_type = AbsBufferComplement,
						logger = self.logger
					)
				for conf_el in self.complement
			]
		else:
			self.complementers = None


	def supply(self, t3s: T3Store) -> Generator[AmpelBuffer, None, None]:

		# NB: we consume the entire cursor at once using list() to be robust
		# against cursor timeouts or server restarts during long lived T3 processes
		stock_ids = list(self.selector.fetch() or [])
		if not stock_ids:
			return

		# Usually, id_key is '_id' but it can be 'stock' if the
		# selection is based on t2 documents for example
		id_key = self.selector.field_name

		# Run start
		###########
		chunks = chunks_func(stock_ids, self.chunk_size) if self.chunk_size > 0 else [stock_ids]

		# Loop over chunks from the cursor/iterator
		for chunk_ids in chunks:

			# allow working chunks to complete even if some raise exception
			try:

				# Load info from DB
				tran_data = self.data_loader.load([sid[id_key] for sid in chunk_ids])

				# Potentialy add complementary information (spectra, TNS names, ...)
				if self.complementers:
					for appender in self.complementers:
						appender.complement(tran_data, t3s)

				yield from tran_data

			except Exception as e:  # noqa: PERF203
				self.event_hdlr.handle_error(e, self.logger)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/DistConfigBuilder.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.10.2019
# Last Modified Date:  18.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import re
from collections.abc import Callable, Sequence
from functools import partial
from importlib import metadata
from typing import Any

import yaml

from ampel.config.builder.ConfigBuilder import ConfigBuilder
from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.log.AmpelLogger import SHOUT, VERBOSE, AmpelLogger
from ampel.util.distrib import PathLike, PathList, get_dist_names, \
	get_files, get_classes_info_from_dist, get_classes_ancestry

class DistConfigBuilder(ConfigBuilder):
	"""
	Subclass of ConfigLoader allowing to automatically detect and load
	various configuration files contained in ampel sub-packages installed with pip
	"""

	def __init__(self,
		options: DisplayOptions,
		logger: None | AmpelLogger = None,
		ignore_exc: list[str] | None = None
	) -> None:
		super().__init__(options, logger, ignore_exc)
		self.units_parent: dict[str, list[str]] = {}
		self.units_ancestry: dict[str, list[str]] = {}
		self.discover_distrib_units('ampel-interface') # populates self.units_parent
		self.visited = {'ampel-interface'}


	def load_distributions(self,
		prefixes: Sequence[str] = ("pyampel-", "ampel-"),
		conf_dirs: Sequence[str] = ("conf",),
		exts: Sequence[str] = ("json", "yaml", "yml"),
		raise_exc: bool = True,
		exclude: None | list[str] = None
	) -> None:
		"""
		:param prefixes: loads all known conf files from all distributions with name starting with prefixes
		:param conf_dirs: loads only conf files from these directories
		:param exts: loads only conf files with these extensions
		:param exclude: exclude distribution by name
		"""
		dist_names = get_dist_names(prefixes)

		if dist_names:
			self.logger.log(SHOUT, f"Detected ampel components: {', '.join(repr(name) for name in dist_names)}")

		if exclude:
			for i in range(len(exclude)):
				exclude[i] = exclude[i].lower()

		for dist_name in dist_names:

			self.logger.break_aggregation()

			if exclude and dist_name.lower() in exclude:
				self.logger.log(VERBOSE, "Excluding distribution '{dist_name}' as requested")
				continue

			s = f"Scanning distribution '{dist_name}'"
			self.logger.log(VERBOSE, s)

			for conf_dir in conf_dirs:
				for ext in exts:
					self.load_distrib(dist_name, conf_dir, ext, raise_exc=raise_exc)

			self.add_distrib_units(dist_name)

			if self.verbose:
				self.logger.log(VERBOSE, f"Done scanning distribution '{dist_name}'")

		if self.verbose:
			self.logger.break_aggregation()
			self.logger.log(VERBOSE, f"Done scanning distributions {', '.join(repr(name) for name in dist_names)}")


	def load_distrib(self,
		dist_name: str,
		conf_dir: str = "conf",
		ext: str = "json",
		raise_exc: bool = True
	) -> None:
		"""
		Loads all known conf files of the provided distribution (name)
		"""
		try:

			distrib = metadata.distribution(dist_name)
			all_conf_files = get_files(dist_name, conf_dir, re.compile(rf".*\.{ext}$"))

			"""
			if all_conf_files and self.verbose:
				suffix = "s" if len(all_conf_files) > 1 else ""
				self.logger.log(VERBOSE, f"Parsing following config file{suffix}:")
				for el in all_conf_files:
					self.logger.log(VERBOSE, el.as_posix())
				self.logger.break_aggregation()
			"""

			if ampel_conf := self.get_conf_file(all_conf_files, f"ampel.{ext}"):
				self.load_conf_using_func(
					distrib, ampel_conf, self.load_ampel_conf, raise_exc=raise_exc
				)

			# Channel, mongo (and template) can be defined by multiple files
			# in a directory named after the corresponding config section name
			for key in ("channel", "mongo", "process"):
				if specialized_conf_files := self.get_conf_files(all_conf_files, f"*/{key}/*"):
					for f in specialized_conf_files:
						self.load_conf_using_func(distrib, f, self.first_pass_config[key].add)

			# ("channel", "mongo", "resource")
			for key in self.first_pass_config.conf_keys:
				if key == "alias":
					continue
				if section_conf_file := self.get_conf_file(all_conf_files, f"{key}.{ext}"):
					self.load_conf_using_func(distrib, section_conf_file, self.first_pass_config[key].add)

			# ("controller", "processor", "unit", "alias", "process")
			for unit_type in ("alias", "process"):
				if tier_conf_file := self.get_conf_file(all_conf_files, f"{unit_type}.{ext}"):
					self.load_conf_using_func(
						distrib, tier_conf_file, partial(self.register_tier_conf, unit_type)
					)

			# Try to load templates from folder template (defined by 'Ampel-ZTF' for ex.)
			if template_conf_files := self.get_conf_files(all_conf_files, "*/template/*"):
				for f in template_conf_files:
					self.load_conf_using_func(distrib, f, self.register_templates)

			# Try to load templates from template.conf
			if template_conf := self.get_conf_file(all_conf_files, "template.{ext}"):
				self.load_conf_using_func(distrib, template_conf, self.register_templates)

			if all_conf_files:
				self.logger.info(f"Not all config files were loaded from distribution '{distrib.name}'")
				suffix = "s" if len(all_conf_files) > 1 else ""
				self.logger.info(f"Unprocessed config file{suffix}: {all_conf_files}")

			self.add_distrib_units(dist_name)

		except Exception as e:
			if raise_exc:
				raise
			self.error = True
			self.logger.error(
				f"Error occured while loading configurations from distribution '{dist_name}'",
				exc_info=e
			)


	def add_distrib_units(self, dist_name: str) -> None:

		if dist_name in self.visited:
			return

		self.visited.add(dist_name)

		units = self.discover_distrib_units(dist_name)

		if self.verbose and units:
			self.logger.log(VERBOSE, "Discovered unit" + "s:" if len(units) > 1 else ":")
			for el in units:
				self.logger.log(VERBOSE, f"{el}")

		self.first_pass_config['unit'].add(
			["ampel." + k.split("ampel/")[1].replace("/", ".").removesuffix(".py") for k in units],
			dist_name, metadata.distribution(dist_name).version, list(units.keys())
		)


	def discover_distrib_units(self,
		dist_name: str,
		required_parents: Sequence[str] = ['AmpelABC'],
		exclude_parents: Sequence[str] = (
			'Secret', 'AbsCLIOperation', 'ConfigCollector',
			'AbsForwardConfigCollector', 'NonDiscoverableUnit'
		)
	) -> dict[str, list[str]]:
		"""
		Discover and filters unit classes from a distribution.
		Updates internal parents and ancestry maps
		"""
		dist_classes_info = get_classes_info_from_dist(dist_name)
		self.units_parent |= dist_classes_info

		get_classes_ancestry(
			self.units_parent,
			self.units_ancestry,
			required_parents = required_parents,
			exclude_parents = exclude_parents
		)

		"""
		for k in dist_classes_info:
			if k in self.units_ancestry:
				print(k, self.units_ancestry[k])
		"""

		return {k: dist_classes_info[k] for k in dist_classes_info if k in self.units_ancestry}


	def register_tier_conf(self,
		root_key: str,
		d: dict[str, Any],
		dist_name: str,
		version: str,
		file_rel_path: str,
	):
		for k in ("t0", "t1", "t2", "t3", "ops"):
			if k in d:
				self.first_pass_config[root_key][k].add(
					d[k],
					dist_name = dist_name,
					version = version,
					register_file = file_rel_path,
				)


	def load_conf_using_func(self,
		distrib: metadata.Distribution,
		file_rel_path: "PathLike",
		func: Callable[[dict[str, Any], str, str, str], None],
		raise_exc: bool = True,
		**kwargs
	) -> None:

		try:

			if self.verbose:
				self.logger.log(VERBOSE, f"Reading {file_rel_path}")

			# NB: read both YAML and JSON (which is a subset of YAML)
			content = yaml.safe_load(file_rel_path.read_text())

			func(
				content,
				distrib.name,
				distrib.version,
				file_rel_path.as_posix(),
				**kwargs
			)

		except FileNotFoundError:
			if raise_exc:
				raise
			self.error = True
			self.logger.error(
				f"File '{file_rel_path}' not found in distribution '{distrib.name}'"
			)

		except Exception as e:
			if raise_exc:
				raise
			self.error = True
			self.logger.error(
				f"Error occured loading '{file_rel_path}' from distribution '{distrib.name}'",
				exc_info=e
			)


	@staticmethod
	def get_conf_file(files: "PathList", key: str) -> "None | PathLike":
		"""
		Extract the first entry who matches the provided 'key' from the provided list
		Note: this method purposely modifies the input list (removes matched element)
		"""
		for el in files:
			if el.match(key):
				files.remove(el)
				return el
		return None


	@staticmethod
	def get_conf_files(files: "PathList", key: str) -> "PathList":
		"""
		Extract all entries who matches the provided 'key' from the provided list
		Note: this method purposely modifies the input list (removes matched elements)
		"""
		ret = [el for el in files if el.match(key)]
		for el in ret:
			files.remove(el)
		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/core/abstract/AbsTargetSource.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                15.08.2018
# Last Modified Date:  15.09.2018
# Last Modified By:    jvs

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod


class AbsTargetSource(AmpelABC, abstract=True):
	""" Provides target fields for follow-up searches """

	@abstractmethod
	async def get_targets(self):
		"""
		:yields: a tuple ((ra,dec), radius, (date_min, date_max), [channels])
		"""



import io
import os
import subprocess
import tempfile
from contextlib import nullcontext
from pathlib import Path

import pytest
import yaml
from pydantic import ValidationError
from pytest_mock import MockerFixture

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.base.BadConfig import BadConfig
from ampel.config.builder.ConfigChecker import ConfigChecker
from ampel.config.builder.ConfigValidator import ConfigValidator
from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
from ampel.core.UnitLoader import UnitLoader
from ampel.test.test_JobCommand import run
from ampel.util.mappings import set_by_path


def test_build_config():
    tmp_file = os.path.join(tempfile.mkdtemp(), "ampel_conf.yml")
    io.BytesIO(
        subprocess.check_output(
            [
                "ampel",
                "config",
                "build",
                "--distributions",
                "ampel-interface",
                "ampel-core",
                "-out",
                tmp_file,
            ]
        )
    )
    with open(tmp_file) as f:
        config = yaml.safe_load(f)
    assert ConfigValidator(config).validate() == config


def test_ConfigChecker(testing_config, monkeypatch):
    """
    ConfigValidator validates units without calling their __init__ methods
    """
    with open(testing_config) as f:
        config = yaml.safe_load(f)
    # validates as-is
    checker = ConfigChecker(config)
    checker.validate(raise_exc=True)

    # add a processor with side-effects
    class SideEffect(RuntimeError): ...

    class SideEffectLadenProcessor(AbsEventUnit):
        required: int  # type: ignore[annotation-unchecked]

        def __init__(self, **kwargs) -> None:
            super().__init__(**kwargs)
            raise SideEffect

        def proceed(self, event_hdlr): ...

    config["process"]["t0"]["BadProcess"] = {
        "name": "BadProcess",
        "version": 0,
        "active": True,
        "schedule": "super",
        "processor": {"unit": "SideEffectLadenProcessor", "config": {"required": 1}},
    }
    checker = ConfigChecker(config)

    def get_class_by_name(name, *args, **kwargs):
        if name == "SideEffectLadenProcessor":
            return SideEffectLadenProcessor
        return UnitLoader.get_class_by_name(checker.loader, name, *args, **kwargs)

    monkeypatch.setattr(checker.loader, "get_class_by_name", get_class_by_name)

    # ConfigChecker attempts to instantiate the unit
    with pytest.raises(SideEffect):
        assert checker.validate(raise_exc=True)

    # ConfigValidator just validates the model
    checker = ConfigValidator(config)
    monkeypatch.setattr(checker.loader, "get_class_by_name", get_class_by_name)
    checker.validate(raise_exc=True)

    # ConfigValidator fails if the config does not satisfy the model
    checker.config["process"]["t0"]["BadProcess"]["processor"]["config"].clear()
    with pytest.raises(BadConfig):
        checker.validate()


@pytest.mark.parametrize(
    ("patch", "result"),
    [({}, None), ({"channel.LONG_CHANNEL.purge": {}}, ValidationError)],
)
def test_validate_config(testing_config, tmpdir, patch, result):
    """Validate validates config"""
    tmp_config = tmpdir / "patched_config.yml"
    with testing_config.open() as f:
        config = yaml.safe_load(f)
    for path, item in patch.items():
        set_by_path(config, path, item)
    tmp_config.write_text(yaml.dump(config), "utf-8")
    with pytest.raises(result) if result else nullcontext():
        assert (
            run(
                [
                    "ampel",
                    "config",
                    "validate",
                    "--file",
                    str(tmp_config),
                ]
            )
            == result
        )


def test_collect_bad_unit(tmp_path: Path, mocker: MockerFixture) -> None:
    cb = DistConfigBuilder(DisplayOptions(verbose=True, debug=True))

    bad_unit_config = tmp_path / "unit.yaml"
    with bad_unit_config.open("w") as f:
        yaml.dump(["flerhherher.Floop"], f)

    mocker.patch(
        "ampel.config.builder.DistConfigBuilder.get_files",
        return_value=[bad_unit_config],
    )

    cb.load_distributions()
    assert cb.first_pass_config.has_nested_error()

    with pytest.raises(
        ValueError,
        match=
            r".*Error were reported while gathering configurations \(first pass config\).*",
    ):
        cb.build_config(stop_on_errors=2)



#!/usr/bin/env python
# File:                Ampel-core/ampel/ingest/IngestionWorker.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                19.03.2025
# Last Modified Date:  19.03.2025
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

import signal

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsIngester import AbsIngester
from ampel.abstract.AbsWorker import stop_on_signal
from ampel.core.EventHandler import EventHandler
from ampel.log import AmpelLogger, LogFlag
from ampel.model.UnitModel import UnitModel
from ampel.queue.AbsConsumer import AbsConsumer
from ampel.t2.T2QueueWorker import QueueItem


class IngestionWorker(AbsEventUnit):
    consumer: UnitModel
    ingester: UnitModel = UnitModel(unit="MongoIngester")

    def proceed(self, event_hdlr: EventHandler) -> int:
        """:returns: number of messages processed"""

        run_id = event_hdlr.get_run_id()

        # Loop variables
        doc_counter = 0

        with(
            AmpelLogger.from_profile(
                self.context,
                self.log_profile,
                run_id,
                base_flag=LogFlag.CORE | self.base_log_flag,
            ) as logger,
            stop_on_signal(
                [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGHUP],
                logger,
            ) as stop_token,
            self.context.loader.new(
                self.consumer,
                stop=stop_token,
                unit_type=AbsConsumer,
            ) as consumer,
            self.context.loader.new_context_unit(
                self.ingester,
                context=self.context,
                run_id=run_id,
                tier=-1,
                process_name=self.process_name,
                error_callback=stop_token.set,
                acknowledge_callback=consumer.acknowledge,
                logger=logger,
                sub_type=AbsIngester,
            ) as ingester,
        ):

            # Process docs until next() returns None (breaks condition below)
            while not stop_token.is_set():
                item: None | QueueItem = consumer.consume()

                # No match
                if item is None:
                    if not stop_token.is_set():
                        logger.log(LogFlag.SHOUT, "No more docs to process")
                    break

                doc_counter += 1

                with ingester.group([item]):
                    for stock in item["stock"]:
                        ingester.stock.ingest(stock)
                    for dp in item["t0"]:
                        ingester.t0.ingest(dp)
                    for t1 in item["t1"]:
                        ingester.t1.ingest(t1)
                    for t2 in item["t2"]:
                        ingester.t2.ingest(t2)

        event_hdlr.add_extra(docs=doc_counter)

        return doc_counter




import pytest

from ampel.abstract.AbsOpsUnit import AbsOpsUnit
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.AmpelUnit import AmpelUnit
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.base.LogicalUnit import LogicalUnit
from ampel.core.ContextUnit import ContextUnit
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.model.ingest.CompilerOptions import CompilerOptions
from ampel.model.UnitModel import UnitModel
from ampel.secret.AmpelVault import AmpelVault
from ampel.secret.DictSecretProvider import DictSecretProvider
from ampel.secret.NamedSecret import NamedSecret
from ampel.secret.Secret import Secret
from ampel.test.dummy import DummyUnitResultAdapter


@pytest.fixture
def secrets():
    return DictSecretProvider(
        {
            "dict": {"a": 1, "b": "foo"},
            "str": "blahblablah",
            "tuple": (1, 1),
        }
    )

@pytest.fixture
def secret_context(dev_context: DevAmpelContext, secrets, monkeypatch):
    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))
    return dev_context


@pytest.mark.parametrize(
    ("expected_type","key"), [(dict, "dict"), (str, "str"), (tuple, "tuple")]
)
def test_resolve_secrets_correct_type(
    dev_context: DevAmpelContext,
    secrets: DictSecretProvider,
    monkeypatch,
    expected_type,
    key,
    ampel_logger,
):
    @dev_context.register_unit
    class Modelo(LogicalUnit):
        seekrit: NamedSecret[expected_type] = NamedSecret[expected_type](label=key)

    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))

    unit = dev_context.loader.new(
        UnitModel(unit="Modelo"), logger=ampel_logger, unit_type=Modelo
    )
    assert unit.seekrit.get() == secrets.store[key]


@pytest.mark.parametrize(
    ("expected_type","key"), [(dict, "str"), (str, "tuple"), (tuple, "dict")]
)
def test_resolve_secrets_wrong_type(
    secrets, dev_context: DevAmpelContext, monkeypatch, expected_type, key, ampel_logger
):
    @dev_context.register_unit
    class Modelo(LogicalUnit):
        seekrit: NamedSecret[expected_type] = NamedSecret[expected_type](label=key)
    
    vault = AmpelVault(providers=[secrets])
    s = Modelo(logger=ampel_logger).seekrit
    with pytest.raises(ValueError, match="Secret not yet resolved"):
        s.get()
    assert vault.resolve_secret(s, expected_type) is False
    
    monkeypatch.setattr(dev_context.loader, "vault", vault)
    with pytest.raises(TypeError, match=f"Could not resolve secret '{key}' as {expected_type.__name__:s}"):
        dev_context.loader.new(
            UnitModel(unit="Modelo"), logger=ampel_logger, unit_type=Modelo
        )

def test_resolve_secret_in_union(secret_context: DevAmpelContext, ampel_logger):
    """UnitLoader can resolve secret in a union field"""

    @secret_context.register_unit
    class Modelo(LogicalUnit):
        maybe_secret: None | NamedSecret[str] = NamedSecret[str](label="nonesuch")

    unit = secret_context.loader.new(
        UnitModel(unit="Modelo", config={"maybe_secret": {"label": "str"}}), logger=ampel_logger, unit_type=Modelo
    )
    assert unit.maybe_secret is not None
    assert unit.maybe_secret.get() is not None

    with pytest.raises(TypeError, match="Could not resolve secret 'nonesuch' as str"):
        secret_context.loader.new(
            UnitModel(unit="Modelo"), logger=ampel_logger, unit_type=Modelo
        )

    unit = secret_context.loader.new(
        UnitModel(unit="Modelo", config={"maybe_secret": None}), logger=ampel_logger, unit_type=Modelo
    )
    assert unit.maybe_secret is None


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_resolve_secret_untyped_default(secret_context: DevAmpelContext, ampel_logger):
    """UnitLoader can resolve secret where default is missing a type parameter"""

    @secret_context.register_unit
    class Modelo(LogicalUnit):
        maybe_secret: NamedSecret[str] = NamedSecret(label="str")

    unit = secret_context.loader.new(
        UnitModel(unit="Modelo"), logger=ampel_logger, unit_type=Modelo
    )
    assert unit.maybe_secret is not None
    assert unit.maybe_secret.get() is not None


def test_resolve_secret_from_superclass(
    secrets: DictSecretProvider, dev_context: DevAmpelContext, monkeypatch, ampel_logger
):
    """Secrets are resolved with multiple inheritance"""

    class Base(ContextUnit):
        seekrit: NamedSecret[dict]

    @dev_context.register_unit
    class Derived(AbsOpsUnit, Base):
        def run(self, beacon):
            ...

    assert (
        "seekrit" not in Derived.__annotations__
    ), "multiply-inherited AmpelBaseModels are missing annotations"

    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))
    unit = dev_context.loader.new(
        UnitModel(unit="Derived", config={"seekrit": {"label": "dict"}}),
        logger=ampel_logger,
        context=dev_context,
        unit_type=Derived,
    )
    assert unit.seekrit.get() == secrets.store.get("dict")


def test_resolve_secret_from_config(
    secrets, dev_context: DevAmpelContext, monkeypatch, ampel_logger
):
    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))

    @dev_context.register_unit
    class NiceAndConcrete(LogicalUnit):
        seekrit: NamedSecret[dict]

    # unit with concrete secret field can be instantiated
    dev_context.loader.new(
        UnitModel(unit="NiceAndConcrete", config={"seekrit": {"label": "dict"}}),
        logger=ampel_logger,
        unit_type=NiceAndConcrete,
    )

    # and also validated without instantiating
    with dev_context.loader.validate_unit_models():
        UnitModel(unit="NiceAndConcrete", config={"seekrit": {"label": "dict"}})
        with pytest.raises(TypeError):
            UnitModel(unit="NiceAndConcrete")

    # unit with abstract secret field cannot be instantiated
    @dev_context.register_unit
    class BadAndAbstract(LogicalUnit):
        seekrit: Secret[dict]

    with pytest.raises(TypeError):
        dev_context.loader.new(
            UnitModel(unit="BadAndAbstract", config={"seekrit": {"label": "dict"}}),
            logger=ampel_logger,
            unit_type=BadAndAbstract,
        )


@pytest.mark.parametrize("config", [None, {"seekrit": {"label": "dict"}}])
def test_use_secret_in_init(
    secrets: DictSecretProvider,
    dev_context: DevAmpelContext,
    monkeypatch,
    ampel_logger,
    config,
):
    """Secrets are populated before being passed to init"""
    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))

    @dev_context.register_unit
    class NeedsSecretInInit(LogicalUnit):
        seekrit: NamedSecret[dict] = NamedSecret[dict](label="dict")

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            assert (
                self.seekrit.get() == secrets.store[self.seekrit.label]
            ), "secret is populated"
    
    with pytest.raises(ValueError, match="Secret not yet resolved"):
        NeedsSecretInInit(logger=ampel_logger).seekrit.get()

    # secret field is populated
    unit = dev_context.loader.new(
        UnitModel(unit="NeedsSecretInInit", config=config),
        logger=ampel_logger,
        unit_type=NeedsSecretInInit,
    )
    assert unit.seekrit.get() == secrets.store["dict"]


def test_secret_in_nested_model(secret_context: DevAmpelContext, ampel_logger):
    """Secrets are resolved for UnitModels embedded in other models"""

    @secret_context.register_unit
    class Outer(LogicalUnit):
        inner: UnitModel
        # include a secret at the top level to ensure it is not resolved twice
        outer_secret: NamedSecret[dict] = NamedSecret[dict](label="dict")

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            # Common pattern: instantiate an aux unit from a model in __init__
            self.inner_thing = AuxUnitRegister.new_unit(model=self.inner)

    @secret_context.register_unit
    class Inner(AmpelUnit):
        # this nested secret should be resolved
        seekrit: NamedSecret[str] = NamedSecret[str](label="str")

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            assert self.seekrit.get() == "blahblablah"

    class SecretModel(AmpelBaseModel):
        seekrit: NamedSecret[str] = NamedSecret[str](label="str")

    @secret_context.register_unit
    class InnerModel(AmpelUnit):
        model: SecretModel

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            assert self.model.seekrit.get() == "blahblablah"

    assert (
        secret_context.loader.new(
            UnitModel(unit="Outer", config={"inner": {"unit": "Inner"}}),
            logger=ampel_logger,
            unit_type=Outer,
        ).inner_thing.seekrit.get()
        == "blahblablah"
    ), "loading unit resolves secret"

    assert (
        secret_context.loader.new(
            UnitModel(
                unit="Outer",
                config={
                    "inner": {
                        "unit": "InnerModel",
                        "config": {"model": {"seekrit": {"label": "str"}}},
                    }
                },
            ),
            logger=ampel_logger,
            unit_type=Outer,
        ).inner_thing.model.seekrit.get()
        == "blahblablah"
    ), "loading unit resolves secret"


def test_unit_validation(dev_context: DevAmpelContext):
    @dev_context.register_unit
    class Dummy(LogicalUnit):
        param: int = 42

    with dev_context.loader.validate_unit_models():
        # simple, one-level validation
        UnitModel(unit="Dummy")
        UnitModel(unit="Dummy", config={"param": 37})
        with pytest.raises(TypeError):
            UnitModel(unit="Dummy", config={"param": "fish"})
        with pytest.raises(TypeError):
            UnitModel(unit="Dummy", config={"nonexistant_param": True})

        t3_config = dict(
            supply = {
                "unit": "T3DefaultBufferSupplier",
                "config": {
                    "select": {"unit": "T3StockSelector"},
                    "load": {
                        "unit": "T3SimpleDataLoader",
                        "config": {"directives": [{"col": "stock"}]},
                    },
                },
            },
            stage = {
                "unit": "T3SimpleStager",
                "config": {"execute": [{"unit": "DemoT3Unit"}]},
            }
        )

        # recursive validation
        UnitModel(unit="T3Processor", config=t3_config)

        t3_config["supply"]["config"]["select"]["unit"] = "NotActuallyAUnit" # type: ignore[index]
        with pytest.raises(TypeError, match=r".*Ampel unit not found: NotActuallyAUnit.*"):
            UnitModel(unit="T3Processor", config=t3_config)


def test_compiler_options_validation(mock_context: DevAmpelContext):
    """AuxAliasableUnit can be intialized from a string"""

    @mock_context.register_unit
    class Dummy(LogicalUnit):
        compiler_options: None | CompilerOptions

    with mock_context.loader.validate_unit_models():
        UnitModel(unit="Dummy")
        UnitModel(unit="Dummy", config={"compiler_options": "DummyCompilerOptions"})
        with pytest.raises(TypeError):
            UnitModel(unit="Dummy", config={"compiler_options": "foo"})


def test_secret_validation(secret_context: DevAmpelContext):
    @secret_context.register_unit
    class Dummy(LogicalUnit):
        seekrit: NamedSecret[str] = NamedSecret[str](label="foo")

    with secret_context.loader.validate_unit_models():
        # secret does not exist
        with pytest.raises(TypeError):
            UnitModel(unit="Dummy")
        # exists, but wrong type
        with pytest.raises(TypeError):
            UnitModel(unit="Dummy", config={"seekrit": {"label": "dict"}})
        # exists with correct type
        UnitModel(unit="Dummy", config={"seekrit": {"label": "str"}})


def test_result_adapter_trace(mock_context: DevAmpelContext):

    mock_context.register_unit(DummyUnitResultAdapter)
    model = UnitModel(unit="DummyUnitResultAdapter")
    u1 = mock_context.loader.new_context_unit(model, mock_context, run_id=1)
    u2 = mock_context.loader.new_context_unit(model, mock_context, run_id=2)
    assert (
        u1._get_trace_content() == u2._get_trace_content()
    ), "trace content is identical for different run_id"
    assert u1._trace_id == u2._trace_id, "trace id is identical for different run_id"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsConfigMorpher.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                27.10.2019
# Last Modified Date:  05.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any

from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.decorator import abstractmethod
from ampel.log.AmpelLogger import AmpelLogger


class AbsConfigMorpher(AmpelABC, AmpelBaseModel, abstract=True):
	"""
	Template class aiming at forging processes conforming with ProcessModel
	"""

	template: None | str

	@abstractmethod
	def morph(self, ampel_config: dict[str, Any], logger: AmpelLogger) -> dict[str, Any]:
		raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/aux/SimpleTagFilter.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                06.12.2021
# Last Modified Date:  06.12.2021
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

from ampel.abstract.AbsApplicable import AbsApplicable
from ampel.content.DataPoint import DataPoint
from ampel.types import Tag


class SimpleTagFilter(AbsApplicable):

    #: Accept DataPoints with any of these tags
    require: None | list[Tag] = None
    #: Reject Datapoints with any of these tags
    forbid: None | list[Tag] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._allow = None if self.require is None else set(self.require)
        self._deny = None if self.forbid is None else set(self.forbid)

    def _accept(self, dp: DataPoint):
        tag = set(dp.get("tag", []))
        return (self._allow is None or tag.intersection(self._allow)) and (
            self._deny is None or not tag.intersection(self._deny)
        )

    def apply(self, arg: list[DataPoint]) -> list[DataPoint]:
        return [el for el in arg if self._accept(el)]
